import ctypes
import gettext
import locale
import logging
import os
import sys

from .. import PluginBase


LOGGER = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self):
        super().__init__()
        self.libintl = None

    def get_interfaces(self):
        return [
            'l10n',
            'l10n_init',
        ]

    def get_deps(self):
        return [
            {
                # if frozen, we need sys._MEIPASS to be set correctly
                'interface': 'frozen',
                'defaults': ['openpaperwork_core.frozen'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'resources',
                'defaults': ['openpaperwork_core.resources.setuptools'],
            },
        ]

    def init(self, core):
        super().init(core)

        if os.name == "nt" and os.getenv('LANG') is None:
            (lang, enc) = locale.getdefaultlocale()
            os.environ['LANG'] = lang

        try:
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            # happens, for instance when LC_ALL is set to a nonexisting locale
            LOGGER.warning(
                "Failed to set localization. Localization will be disabled"
            )
            return

        self.libintl = None
        if getattr(sys, 'frozen', False):
            libintl_path = os.path.abspath(os.path.join(
                sys._MEIPASS, "libintl-8.dll"
            ))
            if not os.path.exists(libintl_path):
                LOGGER.warning("Failed to find library libintl-8.dll")
            else:
                self.libintl = ctypes.cdll.LoadLibrary(libintl_path)

        self.l10n_load('openpaperwork_core.l10n', 'openpaperwork_core')

    def l10n_load(self, python_package, text_domain):
        path_ctx = self.core.call_success(
            "resources_get_dir", python_package, 'out'
        )
        with path_ctx as path:
            if path is None:
                LOGGER.error(
                    "Failed to access ressources '%s/out'", python_package
                )
                return None

            path = str(path)

            mo_file = gettext.find(text_domain, path)
            if mo_file is None:
                # expected if we don't have translation for the user language
                LOGGER.info(
                    "Failed to find valid locale for '%s' (path=%s)",
                    text_domain, path
                )
                # we still try to keep going

            LOGGER.info("Binding text domain %s to '%s'", text_domain, path)

            if self.libintl is not None:
                self.libintl.bindtextdomain(text_domain, path)
                self.libintl.bind_textdomain_codeset(text_domain, 'UTF-8')

            for module in (gettext, locale):
                if hasattr(module, 'bindtextdomain'):
                    module.bindtextdomain(text_domain, path)
                if hasattr(module, 'textdomain'):
                    module.textdomain(text_domain)



import logging
import random

import openpaperwork_core
import openpaperwork_core.promise

from .. import _

LOGGER = logging.getLogger(__name__)

LABELS_FILENAME = "labels"


COLOR_NAMES = [
    ('aqua', (0, 1, 1)),
    ('black', (0, 0, 0)),
    ('blue', (0, 0, 1)),
    ('fuchsia', (1, 0, 1)),
    ('gray', (0x80 / 0xFF, 0x80 / 0xFF, 0x80 / 0xFF)),
    ('green', (0, 0x80 / 0xFF, 0)),
    ('lime', (0, 1, 0)),
    ('maroon', (0x80 / 0xFF, 0, 0)),
    ('navy', (0, 0, 0x80 / 0xFF)),
    ('olive', (0x80 / 0xFF, 0x80 / 0xFF, 0)),
    ('purple', (0x80 / 0xFF, 0, 0x80 / 0xFF)),
    ('red', (1, 0, 0)),
    ('silver', (0xC0 / 0xFF, 0xC0 / 0xFF, 0xC0 / 0xFF)),
    ('teal', (0, 0x80 / 0xFF, 0x80 / 0xFF)),
    ('white', (1, 1, 1)),
    ('yellow', (1, 1, 0)),
]


class LabelLoader(object):
    """
    Go through all the documents to figure out what labels exist.
    """
    def __init__(self, plugin):
        self.plugin = plugin
        self.core = plugin.core
        self.all_docs = []

    def get_promise(self):
        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self.core.call_all,
            args=("storage_get_all_docs", self.all_docs)
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_all, "on_label_loading_start",
        )
        promise = promise.then(
            openpaperwork_core.promise.ThreadedPromise(
                self.core, self.load_labels
            )
        )
        promise = promise.then(lambda *args, **kwargs: None)
        promise = promise.then(
            self.core.call_all, "on_label_loading_end"
        )

        return promise

    def load_labels(self, *args, **kwargs):
        nb_docs = len(self.all_docs)

        LOGGER.info("Loading labels from %d documents", nb_docs)
        for (doc_idx, (doc_id, doc_url)) in enumerate(self.all_docs):
            self.core.call_all(
                "on_progress", "label_loading",
                doc_idx / nb_docs,
                _("Loading labels of document {}").format(doc_id)
            )
            labels = set()
            self.plugin.doc_get_labels_by_url(labels, doc_url)
            for label in labels:
                self.plugin.all_labels[label[0]] = label[1]
        self.core.call_all("on_progress", "label_loading", 1.0)
        LOGGER.info(
            "%d labels loaded from %d documents",
            len(self.plugin.all_labels), nb_docs
        )


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        # {label_name: color, label_name: color, ...}
        self.all_labels = {}

    def get_interfaces(self):
        return [
            "doc_labels",
            "syncable",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def doc_internal_get_mtime_by_url(self, out: list, doc_url):
        labels_url = self.core.call_success(
            "fs_join", doc_url, LABELS_FILENAME
        )
        if self.core.call_success("fs_exists", labels_url) is None:
            return
        out.append(self.core.call_success("fs_get_mtime", labels_url))

    def doc_has_labels_by_url(self, doc_url):
        labels_url = self.core.call_success(
            "fs_join", doc_url, LABELS_FILENAME
        )
        return self.core.call_success("fs_exists", labels_url)

    def doc_get_labels_by_url(self, out: set, doc_url):
        labels_url = self.core.call_success(
            "fs_join", doc_url, LABELS_FILENAME
        )
        if self.core.call_success("fs_exists", labels_url) is None:
            return
        with self.core.call_success("fs_open", labels_url, 'r') as file_desc:
            for line in file_desc.readlines():
                line = line.strip()
                if line == "":
                    continue
                # Expected: ('label', '#rrrrggggbbbb')
                out.add(tuple(x.strip() for x in line.split(",", 1)))

    def doc_get_labels_by_url_promise(self, out: list, doc_url):
        def get_labels(labels=None):
            if labels is None:
                labels = set()
            self.doc_get_labels_by_url(labels, doc_url)
            return labels

        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, get_labels
        )
        out.append(promise)

    def label_generate_color(self):
        color = (
            random.randint(0, 255) / 255,
            random.randint(0, 255) / 255,
            random.randint(0, 255) / 255,
        )
        return self.label_color_from_rgb(color)

    def label_get_foreground_color(self, bg_color):
        brightness = (
            (bg_color[0] * 0.299)
            + (bg_color[1] * 0.587)
            + (bg_color[2] * 0.114)
        )
        if brightness > (69 / 255):
            return (0, 0, 0)  # black
        else:
            return (1, 1, 1)  # white

    def doc_add_label_by_url(self, doc_url, label, color=None):
        assert "," not in label
        current = set()
        self.doc_get_labels_by_url(current, doc_url)
        current = {k: v for (k, v) in current}
        if label in current:
            LOGGER.warning(
                "Label '%s' already on document '%s'", label, doc_url
            )
            return

        if color is not None:
            self.all_labels[label] = color
        if label in self.all_labels:
            color = self.all_labels[label]
        else:
            color = self.label_generate_color()

        LOGGER.info("Adding label '%s' on document '%s'", label, doc_url)

        labels_url = self.core.call_success(
            "fs_join", doc_url, LABELS_FILENAME
        )
        with self.core.call_success("fs_open", labels_url, 'a') as file_desc:
            file_desc.write("{},{}\n".format(label, color))

        if label not in self.all_labels:
            self.all_labels[label] = color

        return True

    def doc_remove_label_by_url(self, doc_url, label):
        LOGGER.info("Removing label '%s' from document '%s'", label, doc_url)

        labels_url = self.core.call_success(
            "fs_join", doc_url, LABELS_FILENAME
        )
        if self.core.call_success("fs_exists", labels_url) is None:
            return

        with self.core.call_success("fs_open", labels_url, 'r') as file_desc:
            labels = file_desc.readlines()

        labels = [lab.strip() for lab in labels]
        labels = [lab.split(",", 1) for lab in labels if len(lab) > 0]
        labels = {l: c for (l, c) in labels}
        try:
            labels.pop(label)
        except KeyError:
            LOGGER.warning(
                "Tried to remove label '%s' from document '%s', but label"
                " was not found on the document", label, doc_url
            )

        labels = [(label, color) for (label, color) in labels.items()]
        labels.sort()

        with self.core.call_success("fs_open", labels_url, "w") as file_desc:
            for (label, color) in labels:
                file_desc.write("{},{}\n".format(label, color))
        return True

    def labels_get_all(self, out: set):
        for (label, color) in self.all_labels.items():
            out.add((label, color))

    def label_color_to_rgb(self, color):
        if color[0] == '#':
            if len(color) == 13:
                return (
                    int(color[1:5], 16) / 0xFFFF,
                    int(color[5:9], 16) / 0xFFFF,
                    int(color[9:13], 16) / 0xFFFF,
                )
            else:
                return (
                    int(color[1:3], 16) / 0xFF,
                    int(color[3:5], 16) / 0xFF,
                    int(color[5:7], 16) / 0xFF,
                )
        else:
            if color.startswith("rgb("):
                color = color[len("rgb("):-1]
            elif color.startswith("("):
                color = color[len("("):-1]
            color = color.split(",")
            color = tuple([int(x.strip()) for x in color])
            color = (color[0] / 0xFF, color[1] / 0xFF, color[2] / 0xFF)
            return color

    def label_color_from_rgb(self, color):
        return (
            "#"
            + format(int(color[0] * 0xFF), '02x') + "00"
            + format(int(color[1] * 0xFF), '02x') + "00"
            + format(int(color[2] * 0xFF), '02x') + "00"
        )

    def label_color_rgb_to_text(self, color):
        for (name, rgb) in COLOR_NAMES:
            if color == rgb:
                break
        else:
            def color_distance(a, b):
                return (
                    (abs(a[0] - b[0]) ** 2) +
                    (abs(a[1] - b[1]) ** 2) +
                    (abs(a[2] - b[2]) ** 2)
                )

            closest_color = min((
                (color_distance(rgb, color), name)
                for (name, rgb) in COLOR_NAMES
            ))
            name = "~" + closest_color[1]

        return "#{:02X}{:02X}{:02X} ({})".format(
            int(color[0] * 0xFF),
            int(color[1] * 0xFF),
            int(color[2] * 0xFF),
            name
        )

    def label_load_all(self, promises: list):
        self.all_labels = {}
        promise = LabelLoader(self).get_promise()
        # drop the return value of 'call_all'
        promise = promise.then(lambda *args, **kwargs: None)
        promises.append(promise)

    def sync(self, promises: list):
        self.label_load_all(promises)



import itertools
import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_change_labels"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.menu_add = None
        self.menu_remove = None
        self.idx_generator = itertools.count()

    def get_interfaces(self):
        return [
            'action',
            'action_docs',
            'action_docs_properties',
            'chkdeps',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'doc_selection',
                'defaults': ['paperwork_gtk.doc_selection'],
            },
            {
                'interface': 'gtk_doc_properties',
                'defaults': ['paperwork_gtk.mainwindow.docproperties'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._open_editor)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def _open_editor(self, action, parameter):
        selection = set()
        self.core.call_all("doc_selection_get", selection)

        if len(selection) <= 0:
            LOGGER.info("No document selected")
            return

        LOGGER.info("Opening properties for %d documents", len(selection))
        self.core.call_all("open_docs_properties", selection)



import openpaperwork_core


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        self.new_doc = None

    def get_interfaces(self):
        return ['new_doc']

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
        ]

    def config_put(self, opt, value, *args, **kwargs):
        if opt != "workdir":
            return None
        # Bug report 170: When the work directory has been changed,
        # we have to make sure to drop any reference to it so the user doesn't
        # use it by accident.
        # (keep in mind that self.new_doc = (doc_id, doc_url), and doc_url
        # includes the work directory path)
        self.new_doc = None
        return None

    def get_new_doc(self):
        if self.new_doc is not None:
            if self.core.call_success("is_doc", self.new_doc[1]) is None:
                return self.new_doc
        self.new_doc = self.core.call_success("storage_get_new_doc")
        return self.new_doc



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    LAYOUTS = {
        # name: pages per line (columns)
        'paged': 1,
        'grid': 3,
    }
    MAX_PAGES = max(LAYOUTS.values())

    def __init__(self):
        super().__init__()
        self.controllers = {}
        self.widget_tree = None
        self.scroll = None
        self.page_layout = None
        self.pages = []
        self.nb_visible_pages = 0
        self.widget_to_page = {}
        self.page_to_widget = {}
        self.nb_columns = self.MAX_PAGES

        self.zoom = 0.0
        self.layout_name = None
        self.requested_page_idx = 0

        self.active_page_idx = 0
        self.active_doc = (None, None)
        self.active_doc_mtime = None

    def get_interfaces(self):
        return [
            'chkdeps',
            'doc_open',
            'drag_and_drop_destination',
            'gtk_docview',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_zoomable',
                'defaults': [
                    'paperwork_gtk.gesture.zoom',
                    'paperwork_gtk.keyboard_shortcut.zoom',
                ],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.mainwindow.docview", "docview.css"
        )

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview", "docview.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.scroll = self.widget_tree.get_object("docview_scroll")
        self.scroll.get_vadjustment().connect(
            "value-changed", self._on_vscroll_value_changed
        )
        self.scroll.get_vadjustment().connect(
            "changed", self._on_vscroll_changed
        )

        self.overlay = self.widget_tree.get_object("docview_drawingarea")
        self.overlay.connect("draw", self._on_overlay_draw)

        self.page_layout = self.widget_tree.get_object("docview_page_layout")
        self.page_layout.connect(
            "size-allocate", self._on_layout_size_allocate
        )
        self.page_layout.connect("child-activated", self._on_child_activated)

        self.page_layout.connect("drag-motion", self._on_drag_motion)
        self.page_layout.connect("drag-leave", self._on_drag_leave)
        self.page_layout.connect("draw", self._on_draw)

        self.core.call_all(
            "gtk_fix_headerbar_buttons",
            self.widget_tree.get_object("docview_header")
        )

        self.on_mainwindow_fold_change()

        self.widget_tree.get_object("docview_back").connect(
            "clicked",
            lambda *args, **kwargs: (
                self.core.call_all("mainwindow_show", "left")
            )
        )

        self.core.call_all(
            "mainwindow_add", side="right", name="docview", prio=10000,
            header=self.widget_tree.get_object("docview_header"),
            body=self.widget_tree.get_object("docview_body"),
        )

        self.core.call_success(
            "mainloop_schedule", self.core.call_all,
            "on_gtk_docview_init", self
        )

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def _on_child_activated(self, page_layout, child):
        page = self.widget_to_page[child]
        for controller in self.controllers.values():
            controller.on_page_activated(page)

    def docview_set_bottom_margin(self, height):
        self.widget_tree.get_object("docview_padding").set_size_request(
            1, height + 10
        )

    def docview_get_headerbar(self):
        return self.widget_tree.get_object("docview_header")

    def docview_get_body(self):
        return self.widget_tree.get_object("docview_overlay")

    def docview_get_scrollwindow(self):
        return self.widget_tree.get_object("docview_scroll")

    def docview_switch_controller(self, name, new_controller_ctor):
        self.controllers[name].exit()
        new_controller = new_controller_ctor(self)
        LOGGER.info(
            "%s: %s --> %s",
            name, str(self.controllers[name]), str(new_controller)
        )
        self.controllers[name] = new_controller
        new_controller.enter()

    def _on_layout_size_allocate(self, layout, allocation):
        for controller in self.controllers.values():
            controller.on_layout_size_allocate(layout)

    def _on_vscroll_value_changed(self, vadj):
        for controller in self.controllers.values():
            controller.on_vscroll_value_changed(vadj)

    def _on_vscroll_changed(self, vadj):
        for controller in self.controllers.values():
            controller.on_vscroll_changed(vadj)

    def _on_drag_motion(self, layout, drag_context, x, y, time):
        for controller in self.controllers.values():
            controller.on_drag_motion(drag_context, x, y, time)

    def _on_drag_leave(self, layout, drag_context, time):
        for controller in self.controllers.values():
            controller.on_drag_leave(drag_context, time)

    def drag_and_drop_get_destination(self, widget, x, y):
        if widget != self.page_layout:
            return None
        for controller in self.controllers.values():
            r = controller.drag_and_drop_get_destination(x, y)
            if r is not None:
                return r
        return None

    def _on_draw(self, layout, cairo_context):
        for controller in self.controllers.values():
            controller.on_draw(cairo_context)

    def _on_overlay_draw(self, widget, cairo_context):
        for controller in self.controllers.values():
            controller.on_overlay_draw(widget, cairo_context)

    def doc_close(self):
        for page in self.pages:
            page.set_visible(False)
        for controller in self.controllers.values():
            controller.on_close()

    def _build_flow_box_child(self, child):
        widget = Gtk.FlowBoxChild.new()
        widget.set_visible(True)
        widget.set_property('name', 'docview_page')
        widget.set_property('halign', Gtk.Align.CENTER)
        widget.add(child)
        return widget

    def doc_open(self, doc_id, doc_url):
        self.doc_close()

        self.core.call_success(
            "mainwindow_show", side="right", name="docview"
        )

        self.zoom = 0.0
        self.layout_name = 'grid'
        self.active_doc = (doc_id, doc_url)
        self.active_doc_mtime = self.core.call_success(
            "doc_get_mtime_by_url", doc_url
        )
        self.active_page_idx = 0

        self.pages = []
        self.widget_to_page = {}
        self.page_to_widget = {}
        self.nb_visible_pages = 0

        self.controllers = {}
        self.core.call_all(
            "gtk_docview_get_controllers", self.controllers, self
        )
        for controller in self.controllers.values():
            controller.enter()

    def doc_reload(self, doc_id, doc_url):
        if self.active_doc != (doc_id, doc_url):
            LOGGER.info(
                "Reload requested for document %s, but active doc is %s",
                (doc_id, doc_url), self.active_doc
            )
            return
        mtime = self.core.call_success("doc_get_mtime_by_url", doc_url)
        if mtime == self.active_doc_mtime:
            LOGGER.info(
                "Reload for document %s requested, but mtime hasn't changed"
                " (%s)", doc_url, mtime
            )
            return
        self.active_doc_mtime = mtime
        LOGGER.info("Reloading document %s", self.active_doc)
        for controller in self.controllers.values():
            controller.doc_reload()

    def on_page_size_obtained(self, page):
        # page may been wrapped
        page = self.pages[page.page_idx]
        for controller in self.controllers.values():
            controller.on_page_size_obtained(page)

    def on_page_shown(self, page_idx):
        self.active_page_idx = page_idx

        page = self.pages[page_idx]
        widget = self.page_to_widget[page]
        self.page_layout.select_child(widget)

    def doc_goto_previous_page(self):
        self.doc_goto_page(self.active_page_idx - 1)

    def doc_goto_next_page(self):
        self.doc_goto_page(self.active_page_idx + 1)

    def doc_goto_page(self, page_idx):
        LOGGER.info(
            "Going to page %d (nb pages=%d)",
            page_idx, self.nb_visible_pages
        )
        if page_idx < 0:
            page_idx = 0
        if page_idx >= self.nb_visible_pages:
            page_idx = self.nb_visible_pages - 1
        self.requested_page_idx = page_idx
        for controller in self.controllers.values():
            controller.doc_goto_page(page_idx)
        self.core.call_all("on_page_shown", page_idx)

    def docview_set_layout(self, name):
        self.layout_name = name
        for controller in self.controllers.values():
            controller.docview_set_layout(name)

    def docview_set_zoom(self, zoom):
        self.zoom = zoom
        for controller in self.controllers.values():
            controller.docview_set_zoom(zoom)

    def docview_add_page_viewer(self, page, visible=True):
        widget = self._build_flow_box_child(page.widget)
        self.widget_to_page[widget] = page
        self.page_to_widget[page] = widget
        self.pages.append(page)
        if visible:
            self.nb_visible_pages += 1
            self.page_layout.add(widget)

    def docview_hide_page_viewer(self, page):
        # Jflesch> Hiding is not enough. We have to remove the widget
        # otherwise it's still taken into account when computing
        # the layout (keep in mind that homogenous is set to True)
        self.page_layout.remove(self.page_to_widget[page])
        self.nb_visible_pages -= 1

    def docview_show_page_viewer(self, page):
        # XXX(Jflesch): We mess up the order of the widgets here. But since
        # only the scan viewer uses this method at the moment, it's fine.
        self.page_layout.add(self.page_to_widget[page])
        self.nb_visible_pages += 1

    def on_mainwindow_fold_change(self):
        folded = self.core.call_success("mainwindow_get_folded")
        self.widget_tree.get_object("docview_header").set_show_close_button(
            not folded
        )
        self.widget_tree.get_object("docview_back").set_visible(folded)



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class Zoomer(object):
    def __init__(self, zoomable, adjustment):
        self.adjustment = adjustment
        self.gesture = Gtk.GestureZoom.new(zoomable)

        self.ref_adj_value = 1.0
        self.ref_gesture_value = 1.0

        self.gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.gesture.connect("update", self._on_gesture, adjustment)
        self.gesture.connect("end", self._update_refs)
        self.adjustment.connect("value-changed", self._update_refs)

    def _on_gesture(self, gesture, sequence, adjustment):
        scale = gesture.get_scale_delta()
        adj_scale = (self.ref_adj_value * scale / self.ref_gesture_value)
        LOGGER.debug("Zoom gesture: %f --> %f", scale, adj_scale)
        adjustment.set_value(adj_scale)

    def _update_refs(self, *args, **kwargs):
        self.ref_adj_value = self.adjustment.get_value()
        self.ref_gesture_value = self.gesture.get_scale_delta()
        LOGGER.debug("Ref: %f, %f", self.ref_adj_value, self.ref_gesture_value)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        # XXX(Jflesch): Looks like a GObject Introspection bug: if the
        # GtkGesture objects get garbage-collected, the gesture isn't detected
        # anymore. --> We need to keep a reference to those objects as long
        # as we need them.
        self.refs = {}

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_zoomable',
        ]

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_zoomable_widget_new(self, zoomable, adjustment):
        """
        zoomable is the widget on which user can zoom.
        zoom gestures are not applied directly on the widget. They are
        applied on a GtkAdjustment.
        """
        zoomer = Zoomer(zoomable, adjustment)
        self.core.call_all("on_objref_track", zoomer)
        self.refs[zoomable] = zoomer

    def on_zoomable_widget_destroy(self, zoomable, adjustment):
        if zoomable in self.refs:
            self.refs.pop(zoomable)



import datetime
import logging
import time

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

try:
    from . import doc_box
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps
import openpaperwork_gtk.deps

from ... import _


LOGGER = logging.getLogger(__name__)

# GtkListBox doesn't scale well with too many elements
# --> by default we only display 50 documents, and only extend the list
# as needed
NB_DOCS_PER_PAGE = 50


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -100

    def __init__(self):
        super().__init__()
        self.widget_tree = None
        self.doclist = None

        self.scrollbar = None
        self._scrollbar_last_value = -1

        self.docs = []
        self.row_visibles = []
        self.doc_loaded = 0
        self.row_to_doc = {}
        self.docid_to_row = {}
        self.min_doc_box_height = 1

        self.last_date = datetime.datetime(year=1, month=1, day=1)

        self.active_doc = (None, None)
        self.previous_doc = (None, None)

        self.main_actions = []
        self.doc_actions = None

        self.selection_multiple = False

        # the following boolean is a dirty hack to ignore the signal 'toggle'
        # when we toggle the checkbox ourselves (if another plugin
        # calls doc_selection_add() for instance)
        self._toggling = False

    def get_interfaces(self):
        return [
            'chkdeps',
            'doc_actions',
            'doc_open',
            'doc_selection',
            'docs_actions',
            'drag_and_drop_destination',
            'gtk_app_menu',
            'gtk_doclist',
            'screenshot_provider',
            'search_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_drag_and_drop',
                'defaults': ['paperwork_gtk.gesture.drag_and_drop'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_mainwindow',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_widget_flowlayout',
                'defaults': ['paprwork_gtk.widget.flowlayout'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def init(self, core):
        super().init(core)

        self.core.call_success(
            "gtk_load_css",
            "paperwork_gtk.mainwindow.doclist", "doclist.css"
        )

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.doclist", "doclist.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

        self.doc_actions = Gio.Menu.new()

        self.doclist = self.widget_tree.get_object("doclist_listbox")
        self.core.call_all(
            "mainwindow_add", side="left", name="doclist", prio=10000,
            header=self.widget_tree.get_object("doclist_header"),
            body=self.widget_tree.get_object("doclist_body"),
        )
        self.core.call_all(
            "mainwindow_add", side="left", name="doclist_selection_multiple",
            prio=-100000,
            header=self.widget_tree.get_object(
                "doclist_header_selection_multiple"
            ),
            body=None
        )

        self.vadj = self.widget_tree.get_object(
            "doclist_scroll"
        ).get_vadjustment()
        self.vadj.connect("value-changed", self._on_scrollbar_value_changed)

        self.doclist.connect("row-activated", self._on_row_activated)
        self.doclist.connect("drag-motion", self._on_drag_motion)
        self.doclist.connect("drag-leave", self._on_drag_leave)
        self.doclist.connect("size-allocate", self._notify_box_visibility)

        self.widget_tree.get_object("doclist_new_doc").connect(
            "clicked", self._on_new_doc
        )

        self.widget_tree.get_object(
            "doclist_back_to_selection_single"
        ).connect("clicked", lambda button: self.core.call_all(
            "gtk_switch_to_doc_selection_single"
        ))

        self.menu_model = self.widget_tree.get_object("doclist_menu_model")
        self.selection_multiple_menu_model = self.widget_tree.get_object(
            "doclist_selection_multiple_menu_model"
        )

        self.core.call_all("drag_and_drop_page_enable", self.doclist)

        self.core.call_one(
            "mainloop_schedule", self.core.call_all, "on_doclist_initialized"
        )

        self.on_mainwindow_fold_change()

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def on_gtk_initialized(self):
        self.core.call_all(
            "mainwindow_show", side="left", name="doclist"
        )

    def doclist_add(self, widget, vposition):
        body = self.widget_tree.get_object("doclist_body")
        body.add(widget)
        body.reorder_child(widget, vposition)

    def _on_new_doc(self, button):
        self.doc_open_new()

    def doc_open_new(self):
        new_doc = self.core.call_success("get_new_doc")
        self.core.call_all("doc_open", *new_doc)
        self.doclist_show(self.docs, show_new=True)

    def doclist_clear(self):
        self.min_doc_box_height = 1
        self.core.call_all("on_doc_list_clear")

        start = time.time()
        for child in self.doclist.get_children():
            self.doclist.remove(child)
        stop = time.time()

        LOGGER.info(
            "%d documents cleared in %dms",
            self.doc_loaded, (stop - start) * 1000
        )

        self.last_date = datetime.datetime(year=1, month=1, day=1)
        self.doc_loaded = 0
        self.row_to_doc = {}
        self.docid_to_row = {}
        self.row_visibles = []

    def _add_date_box(self, name, txt):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.doclist", name
        )
        widget_tree.get_object("date_label").set_text(txt)
        row = widget_tree.get_object("date_box")
        self.doclist.insert(row, -1)

    def _toggle_doc(self, button, doc_id, doc_url):
        if button.get_active():
            self.core.call_all("doc_selection_add", doc_id, doc_url)
        else:
            self.core.call_all("doc_selection_remove", doc_id, doc_url)

    def _add_doc_box(self, doc_id, doc_url, box="doc_box.glade", new=False):
        row = doc_box.DocBox()

        flowlayout = self.core.call_success(
            "gtk_widget_flowlayout_new", spacing=(3, 3)
        )
        flowlayout.set_visible(True)
        row.layout_bin.add(flowlayout)

        self.core.call_all(
            "on_doc_box_creation", doc_id, row, flowlayout
        )

        doc_actions = row.main_actions
        if new:
            doc_actions.set_visible(False)
        else:
            row.action_menu.set_menu_model(self.doc_actions)

        for action in self.main_actions:
            action = doc_box.DocMainAction(
                action['txt'], action['icon_name'], action['callback']
            )
            row.main_actions.pack_start(
                action, expand=True, fill=True, padding=0
            )

        row.selector.set_visible(
            self.selection_multiple
        )
        row.selector.set_active(
            self.core.call_success("doc_selection_in", doc_id, doc_url)
            is not None
        )
        row.selector.connect(
            "toggled", self._toggle_doc, doc_id, doc_url
        )

        row.connect("size-allocate", self._on_doc_box_size_allocate)

        self.row_to_doc[row] = (doc_id, doc_url)
        self.docid_to_row[doc_id] = row
        self.doclist.insert(row, -1)

    def doclist_extend(self, nb_docs):
        start = time.time()
        # FIXME: show month only when there are too many documents. How many?
        show_month = True

        docs = self.docs[
            self.doc_loaded:self.doc_loaded + nb_docs
        ]
        LOGGER.info(
            "Adding %d documents to the document list (%d-%d)",
            len(docs), self.doc_loaded, self.doc_loaded + nb_docs
        )

        for (doc_id, doc_url) in docs:
            doc_date = self.core.call_success("doc_get_date_by_id", doc_id)

            if doc_date is not None:
                if not show_month:
                    if doc_date.year != self.last_date.year:
                        doc_year = self.core.call_success(
                            "i18n_date_long_year", doc_date
                        )
                        self._add_date_box("year_box.glade", doc_year)

                else:
                    if (doc_date.year != self.last_date.year or
                            doc_date.month != self.last_date.month):
                        doc_year = self.core.call_success(
                            "i18n_date_long_year", doc_date
                        ) + " / " + self.core.call_success(
                            "i18n_date_long_month", doc_date
                        )
                        self._add_date_box("year_box.glade", doc_year)

                self.last_date = doc_date

            self._add_doc_box(doc_id, doc_url)

        self.doc_loaded = min(len(self.docs), self.doc_loaded + nb_docs)

        stop = time.time()

        LOGGER.info(
            "%d documents shown in %dms (%d displayable)",
            len(docs), (stop - start) * 1000, len(self.docs)
        )

        return len(docs)

    def doclist_show(self, docs, show_new=True):
        self.doclist_clear()

        scroll = (self.docs != docs)
        self.docs = docs

        if show_new:
            new_doc = self.core.call_success("get_new_doc")
            self._add_doc_box(*new_doc, new=True)

        self.doclist_extend(NB_DOCS_PER_PAGE)
        self._reselect_current_doc(scroll=scroll)

    def on_search_results(self, query, docs):
        self.doclist_show(docs, show_new=(query == ""))

    def doc_close(self):
        self.active_doc = (None, None)

    def doc_open(self, doc_id, doc_url):
        self.active_doc = (doc_id, doc_url)
        self._reselect_current_doc(scroll=False)

    def _show_doc_id(self, doc_id):
        row = self.docid_to_row.get(doc_id)
        while row is None:
            if self.doclist_extend(NB_DOCS_PER_PAGE) <= 0:
                break
            row = self.docid_to_row.get(doc_id)
        return row

    def _reselect_current_doc(self, scroll=True):
        (doc_id, doc_url) = self.active_doc

        if doc_id not in {doc[0] for doc in self.docs}:
            LOGGER.warning(
                "Document %s not found in the document list", doc_id
            )
            self.vadj.set_value(self.vadj.get_lower())
            return

        row = self._show_doc_id(doc_id)
        assert row is not None
        self.doclist.select_row(row)

        if (self.previous_doc[0] is not None and
                self.previous_doc[0] in self.docid_to_row):
            row = self.docid_to_row[self.previous_doc[0]]
            row.main_actions.set_visible(False)

        if ((not self.selection_multiple) and
                self.core.call_success("is_doc", doc_url) is not None and
                doc_id in self.docid_to_row):
            row = self.docid_to_row[doc_id]
            row.main_actions.set_visible(True)

        self.previous_doc = self.active_doc

        if not scroll:
            return

        handler_id = None

        def scroll_to_row(row, allocation):
            adj = allocation.y
            adj -= self.vadj.get_page_size() / 2
            adj += allocation.height / 2
            min_val = self.vadj.get_lower()
            if adj < min_val:
                adj = min_val
            self.vadj.set_value(adj)
            row.disconnect(handler_id)

        handler_id = row.connect("size-allocate", scroll_to_row)

    def _on_doc_box_size_allocate(self, box, allocation):
        h = allocation.height
        if h <= 0:
            return
        if self.min_doc_box_height <= 1:
            self.min_doc_box_height = h
        else:
            self.min_doc_box_height = min(self.min_doc_box_height, h)
        self._notify_box_visibility()

    def _notify_box_visibility(self, *args, **kwargs):
        lower = int(self.vadj.get_value())
        upper = int(lower + self.vadj.get_page_size())

        row_visibles = []
        prow = None
        for y in range(lower, upper, self.min_doc_box_height):
            row = self.doclist.get_row_at_y(y)
            if row is None or prow is row:
                continue
            prow = row
            row_visibles.append(row)

        if self.row_visibles == row_visibles:
            return
        self.row_visibles = row_visibles

        self.core.call_all("on_doc_list_visibility_changed")
        for row in row_visibles:
            doc = self.row_to_doc.get(row)
            if doc is None:
                continue
            (doc_id, doc_url) = doc
            layout = row.layout_bin.get_child()
            self.core.call_all("on_doc_box_visible", doc_id, row, layout)

    def _on_scrollbar_value_changed(self, vadj):
        lower = vadj.get_lower()
        upper = vadj.get_upper() - lower

        value = (
            vadj.get_value() +
            vadj.get_page_size() -
            lower
        ) / upper

        if value < 0.95:
            self._notify_box_visibility()
            return

        if self._scrollbar_last_value == vadj.get_value():
            # Previous extend call hasn't been taken into account yet
            self._notify_box_visibility()
            return

        self.doclist_extend(NB_DOCS_PER_PAGE)
        self._scrollbar_last_value = vadj.get_value()
        self._notify_box_visibility()

    def _doc_open(self, doc_id, doc_url):
        if self.active_doc[0] == doc_id:
            return
        LOGGER.info("Opening document %s (%s)", doc_id, doc_url)
        self.core.call_all("doc_open", doc_id, doc_url)

    def _on_row_activated(self, list_box, row):
        (doc_id, doc_url) = self.row_to_doc[row]
        self._doc_open(doc_id, doc_url)

    def menu_app_append_item(self, item):
        # they are actually the same menu
        self.doclist_menu_append_item(item)

    def menu_app_append_submenu(self, label, menu):
        # they are actually the same menu
        self.doclist_menu_append_submenu(label, menu)

    def doclist_menu_append_item(self, item):
        self.menu_model.append_item(item)

    def doclist_menu_append_submenu(self, label, menu):
        self.menu_model.append_submenu(label, menu)

    def docs_menu_append_item(self, item):
        self.selection_multiple_menu_model.append_item(item)

    def docs_menu_append_submenu(self, label, menu):
        self.selection_multiple_menu_model.append_submenu(label, menu)

    def add_doc_action(self, action_label, action_name):
        self.doc_actions.append(action_label, action_name)

    def add_doc_main_action(self, icon_name, txt, callback):
        self.main_actions.append({
            "icon_name": icon_name,
            "txt": txt,
            "callback": callback,
        })

    def _on_drag_motion(self, widget, drag_context, x, y, time):
        widget.drag_unhighlight_row()
        row = widget.get_row_at_y(y)
        if row is not None:
            widget.drag_highlight_row(row)

    def _on_drag_leave(self, widget, drag_context, time):
        widget.drag_unhighlight_row()

    def drag_and_drop_get_destination(self, widget, x, y):
        if self.doclist != widget:
            return None
        row = self.doclist.get_row_at_y(y)
        if row is None:
            LOGGER.warning("No row at %d. Can't get drop destination", y)
            return None
        (doc_id, doc_url) = self.row_to_doc[row]
        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            nb_pages = 0
        return (doc_id, doc_url, nb_pages)

    def gtk_open_app_menu(self):
        self.widget_tree.get_object("doclist_menu").clicked()

    def screenshot_snap_app_menu(self, out_file):
        widget = self.widget_tree.get_object("doclist_menu")

        if widget.get_popover() is not None:
            popover = widget.get_popover()
            if popover.get_visible() and popover.is_drawable():
                widget = popover

        self.core.call_success(
            "screenshot_snap_widget",
            widget, out_file,
            margins=(50, 30, 50, 30)
        )

    def screenshot_snap_all_doc_widgets(self, out_dir):
        self.core.call_success(
            "screenshot_snap_widget",
            self.widget_tree.get_object("doclist_new_doc"),
            self.core.call_success("fs_join", out_dir, "doc_new_button.png"),
            margins=(30, 30, 30, 30)
        )

        self.screenshot_snap_app_menu(
            self.core.call_success("fs_join", out_dir, "app_menu.png")
        )

        if self.active_doc[0] is None:
            return
        row = self.docid_to_row[self.active_doc[0]]
        self.core.call_success(
            "screenshot_snap_widget", row.main_actions,
            self.core.call_success(
                "fs_join", out_dir, "doc_properties_button.png"
            ),
            margins=(50, 50, 50, 50)
        )

    def _set_all_selector_visibility(self, visible):
        for row in self.docid_to_row.values():
            checkbox = row.selector
            checkbox.set_visible(visible)

    def gtk_switch_to_doc_selection_single(self):
        self.core.call_all(
            "mainwindow_show", side="left", name="doclist"
        )
        self._set_all_selector_visibility(False)
        self.selection_multiple = False
        self._reselect_current_doc(scroll=False)

    def gtk_switch_to_doc_selection_multiple(self):
        self.core.call_all(
            "mainwindow_show", side="left", name="doclist_selection_multiple"
        )
        if not self.selection_multiple:
            self.core.call_all("doc_selection_reset")
        self.selection_multiple = True
        self._set_all_selector_visibility(True)
        self._reselect_current_doc(scroll=False)

    def doc_selection_reset(self):
        self._update_count()

        self._toggling = True
        try:
            for row in self.docid_to_row.values():
                checkbox = row.selector
                checkbox.set_active(False)
        finally:
            self._toggling = False

    def doc_selection_add(self, doc_id, doc_url):
        self._update_count()

        if self._toggling:
            return
        self._toggling = True
        try:
            row = self.docid_to_row.get(doc_id, None)
            if row is None:
                return
            checkbox = row.selector
            if not checkbox.get_active():
                checkbox.set_active(True)
        finally:
            self._toggling = False

    def doc_selection_remove(self, doc_id, doc_url):
        self._update_count()

        if self._toggling:
            return
        self._toggling = True
        try:
            row = self.docid_to_row.get(doc_id, None)
            if row is None:
                return
            checkbox = row.selector
            if checkbox.get_active():
                checkbox.set_active(False)
        finally:
            self._toggling = False

    def _update_count(self):
        count = self.core.call_success("doc_selection_len")
        if count is None:
            count = 0
        self.widget_tree.get_object(
            "doclist_selection_multiple_nb_docs"
        ).set_text(_("%d documents") % count)

    def open_next_doc(self, offset=1):
        try:
            idx = self.docs.index(self.active_doc)
            idx += offset
        except ValueError:
            idx = 0
        idx = max(idx, 0)
        idx = min(idx, len(self.docs))
        if idx >= len(self.docs):  # may happen with an empty doc list
            return
        (doc_id, doc_url) = self.docs[idx]
        self.core.call_all("doc_open", doc_id, doc_url)
        self._reselect_current_doc(scroll=True)

    def doc_menu_open(self):
        if self.active_doc[0] is None:
            return
        doc_id = self.active_doc[0]
        row = self.docid_to_row[doc_id]
        button = row.action_menu
        button.clicked()

    def screenshot_snap_doc_action_menu(self, out_file):
        if self.active_doc[0] is None:
            return
        doc_id = self.active_doc[0]
        row = self.docid_to_row[doc_id]
        button = row.main_actions

        self.core.call_success(
            "screenshot_snap_widget",
            button,
            out_file,
            margins=(100, 50, 100, 200)
        )

    def on_mainwindow_fold_change(self):
        folded = self.core.call_success("mainwindow_get_folded")
        self.widget_tree.get_object("doclist_header").set_show_close_button(
            folded
        )



"""
To use it:

```sh
paperwork-cli plugins add \
    paperwork_backend.guesswork.label.sklearn.gaussian_with_hashes
paperwork-cli test_sklearn_gaussian_with_hashes
```
"""
import time
import functools

import sklearn.naive_bayes

import openpaperwork_core


samples = 5


class Baye(object):
    def __init__(self, core, label, n_features):
        self.core = core
        self.label = label
        self.data = []
        self.counts = []
        self.tfid = []
        self.targets = []

        self.sklearn_hash_vectorizer = \
            sklearn.feature_extraction.text.HashingVectorizer(
                n_features=n_features
            )
        self.sklearn_tfid_transformer = \
            sklearn.feature_extraction.text.TfidfTransformer(use_idf=False)

        # Best accuracy: ~0.9991
        self.sklearn_classifier = sklearn.naive_bayes.GaussianNB()

    def fit(self, category, text):
        text = text.strip()
        if len(text) <= 0:
            return

        category = (1 if category == 'yes' else 0)
        self.data.append(text)
        self.targets.append(category)

    def seal(self):
        r = 0
        self.core.call_all(
            "on_progress", "fitting", 0.0,
            "Fitting training of label {} ...".format(self.label)
        )
        for pos in range(0, len(self.data), 50):
            self.core.call_all(
                "on_progress", "fitting", pos / len(self.data),
                "Fitting training of label {} ...".format(self.label)
            )
            counts = self.sklearn_hash_vectorizer.fit_transform(
                self.data[pos:pos + 50]
            )
            tfid = self.sklearn_tfid_transformer.fit_transform(
                counts
            ).toarray()

            start = time.time()
            self.sklearn_classifier.partial_fit(
                tfid, self.targets[pos:pos + 50],
                classes=[0, 1]
            )
            stop = time.time()
            r += stop - start
        self.core.call_all("on_progress", "fitting", 1.0)
        return r

    def score(self, text, label):
        global samples

        counts = self.sklearn_hash_vectorizer.transform([text])
        tfid = self.sklearn_tfid_transformer.transform(counts)
        predicted = self.sklearn_classifier.predict_proba(tfid.toarray())
        r = {
            "no": predicted[0][0],
            "yes": predicted[0][1],
        }
        if samples > 0:
            samples -= 1
            print("Prediction sample: {}={}".format(label, r))
        return r


class Bayes(object):
    def __init__(self, core, labels, n_features):
        self.core = core
        self.bayes = {
            label: Baye(core, label, n_features)
            for label in labels
        }

    def fit(self, label, present, text):
        self.bayes[label].fit("yes" if present else "no", text)

    def seal(self):
        r = 0
        for (idx, (k, v)) in enumerate(self.bayes.items()):
            print("Fitting training of label {} ({}/{}) ...".format(
                k, idx, len(self.bayes)
            ))
            r += v.seal()

        s = int(r)
        ms = int(((r) * 1000) % 1000)
        print("Fitting took {}s {}ms for {} labels".format(
            s, ms, len(self.bayes)
        ))

    def score(self, label, text):
        score = self.bayes[label].score(text, label)
        return {
            "yes": score['yes'] if 'yes' in score else 0,
            "no": score['no'] if 'no' in score else 0,
        }


@functools.total_ordering
class Document(object):
    def __init__(self, core, doc_id, text, labels):
        self.core = core
        self.doc_id = doc_id
        self.text = text
        self.labels = labels
        self.scores = {}

    def __lt__(self, other):
        return self.doc_id < other.doc_id

    def __eq__(self, other):
        return self.doc_id == self.doc_id

    def __hash__(self):
        return hash(self.doc_id)

    def fit(self, bayes, all_labels):
        for label in all_labels:
            bayes.fit(label, label in self.labels, self.text)

    def compute_scores(self, bayes, all_labels):
        for label in all_labels:
            self.scores[label] = bayes.score(label, self.text)

    def compute_accuracy(self, all_labels):
        success = 0
        for label in all_labels:
            score = self.scores[label]
            guessed_has_label = score['yes'] > score['no']
            if guessed_has_label == (label in self.labels):
                success += 1
        return success / len(all_labels)


class Corpus(object):
    """
    All the texts of all the documents, and their labels.
    """
    def __init__(self, core):
        self.core = core
        self.documents = []
        self.labels = set()
        self.bayes = None

    def load_all(self):
        docs = []
        self.core.call_all("storage_get_all_docs", docs)
        total = len(docs)
        for (idx, (doc_id, doc_url)) in enumerate(docs):
            self.core.call_all(
                "on_progress", "load_docs",
                idx / total, "Loading document %s" % doc_id
            )
            text = []
            self.core.call_all("doc_get_text_by_url", text, doc_url)
            text = "\n\n".join(text)

            labels = set()
            self.core.call_all("doc_get_labels_by_url", labels, doc_url)
            labels = [label[0] for label in labels]
            labels.sort()
            for label in labels:
                self.labels.add(label)

            self.documents.append(Document(self.core, doc_id, text, labels))

        self.documents.sort()
        self.core.call_all("on_progress", "load_docs", 1.0)

    def mk_bayes(self, n_features):
        self.bayes = Bayes(self.core, self.labels, n_features)

    def fit(self):
        for (idx, doc) in enumerate(self.documents):
            doc.fit(self.bayes, self.labels)

    def seal(self):
        self.bayes.seal()

    def compute_scores(self):
        total = len(self.documents)
        for (idx, doc) in enumerate(self.documents):
            self.core.call_all(
                "on_progress", "scoring",
                idx / total, "computing label scores on %s" % doc.doc_id
            )
            doc.compute_scores(self.bayes, self.labels)
        self.core.call_all("on_progress", "scoring", 1.0)

    def compute_accuracy(self):
        accuracy_sum = 0

        total = len(self.documents)
        for doc in self.documents:
            accuracy_sum += doc.compute_accuracy(self.labels)
        return accuracy_sum / total


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'shell',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'doc_text',
                'defaults': [
                    'paperwork_backend.model.hocr',
                    'paperwork_backend.model.pdf',
                ],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        parser.add_parser('test_sklearn_gaussian_with_hashes')

    def cmd_run(self, console, args):
        if args.command != 'test_sklearn_gaussian_with_hashes':
            return None

        corpus = Corpus(self.core)
        corpus.load_all()

        for two_pow in (16, 18, 20,):
            n_features = 2 ** two_pow
            corpus.mk_bayes(n_features)
            corpus.fit()
            corpus.seal()
            corpus.compute_scores()
            accuracy = corpus.compute_accuracy()

            console.print("n_features=2**{} --> accuracy={}".format(
                two_pow, accuracy
            ))

        return True



from . import PluginBase


class Plugin(PluginBase):
    def get_interfaces(self):
        return ['urls']

    def url_args_join(self, *args, **kwargs):
        out = "".join(args)
        if len(kwargs) <= 0:
            return out
        first = True
        for (k, v) in sorted(list(kwargs.items())):
            if v is None:
                continue
            if first:
                out += "#"
                first = False
            else:
                out += "&"
            out += "{}={}".format(k, v)
        return out

    def url_args_split(self, url):
        if "#" not in url:
            return (url, {})
        (base, kwargs) = url.split("#", 1)
        kwargs = kwargs.split("&")
        out = {}
        for kwarg in kwargs:
            (k, v) = kwarg.split("=", 1)
            out[k] = v
        return (base, out)



"""
Takes care of the few things that must be done if we are running in an
executable created with cx_freeze.

It must be loaded as early as possible.
"""
import multiprocessing
import os
import sys

from . import PluginBase


class Plugin(PluginBase):
    def __init__(self):
        if not getattr(sys, 'frozen', False):
            return
        self._set_meipass()
        multiprocessing.freeze_support()

    def get_interfaces(self):
        return ['frozen']

    def _set_meipass(self):
        # If sys.frozen, then Pyocr (and possibly others) needs MEIPASS to be
        # set *before* importing it.
        if getattr(sys, '_MEIPASS', False):
            # Pyinstaller case
            return
        # Cx_Freeze case
        if "python" not in sys.executable or '__file__' not in globals():
            sys._MEIPASS = os.path.dirname(os.path.realpath(sys.executable))
        else:
            sys._MEIPASS = os.path.realpath(os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "..", ".."
            ))



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps

from ... import _


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -1500

    def get_interfaces(self):
        return [
            'menu',
            'menu_app',
            'menu_app_shortcuts',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_app_open_shortcuts',
                'defaults': ['paperwork_gtk.actions.app.open_shortcuts'],
            },
            {
                'interface': 'gtk_app_menu',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def on_doclist_initialized(self):
        item = Gio.MenuItem.new(_("Shortcuts"), "win.open_shortcuts")
        self.core.call_all("help_add_menu_item", item)



import logging

try:
    from gi.repository import Gio
    HAS_GTK_GLIB = True
except (ImportError, ValueError):
    HAS_GTK_GLIB = False

import openpaperwork_core

from .. import _

LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                'interface': 'gtk_doc_import',
                'defaults': ['paperwork_gtk.docimport'],
            },
            {
                'interface': 'gtk_init',
                'defaults': ['openpaperwork_gtk.gtk_init'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser('import', help=_(
            "Run Paperwork and import files passed as arguments into a new "
            "document"
        ))
        p.add_argument("files", metavar="URLS", type=str, nargs="*", help=_(
            "URLs or paths of files to import"
        ))

    def cmd_run(self, console, args):
        if args.command != 'import':
            return None
        if not HAS_GTK_GLIB:
            LOGGER.error("Cannot import file without gtk and glib")
            return None
        self.core.call_all("gtk_init", self._cmd_run, args)

    def _cmd_run(self, app, options, args):
        app = self.core.call_success("gtk_get_app")

        # when no file is specified, we still start the GUI, because the
        # desktop file always calls paperwork-gtk import (see install.py for
        # the rationale).
        if args.files:
            files = [self.core.call_success("fs_safe", f) for f in args.files]
            giofiles = [Gio.File.new_for_uri(f) for f in files]
            LOGGER.info("Will import files %s", " ".join(files))
            if app.get_is_remote():
                LOGGER.info("Passing control to main paperwork instance")
            # open the files, possibly in another process
            app.open(giofiles, "")
        else:
            # possibly in another process
            if app.get_is_remote():
                LOGGER.info("Passing control to main paperwork instance")
            app.activate()
        return 0



import logging

try:
    import gi
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps


LOGGER = logging.getLogger(__name__)


class LabelingTask(object):
    def __init__(self, plugin, doc_id, doc_url, flowlayout):
        self.plugin = plugin
        self.core = plugin.core

        self.doc_id = doc_id
        self.doc_url = doc_url
        self.flowlayout = flowlayout

    def show_labels(self, labels):
        for widget in list(self.flowlayout.get_children()):
            if hasattr(widget, 'txt'):
                self.flowlayout.remove(widget)

        labels = [
            (
                self.core.call_success("i18n_strip_accents", label[0].lower()),
                label[0],
                label[1]
            ) for label in labels
        ]
        labels.sort()

        for label in labels:
            try:
                color = self.core.call_success("label_color_to_rgb", label[2])
            except Exception as exc:
                LOGGER.warning(
                    "Invalid label %s on document %s", label, self.doc_id,
                    exc_info=exc
                )
                continue
            widget = self.core.call_success(
                "gtk_widget_label_new", label[1], color
            )
            widget.set_visible(True)
            self.flowlayout.add_child(widget, Gtk.Align.END)

    def get_promise(self):
        promise = openpaperwork_core.promise.Promise(
            self.core,
            LOGGER.info, args=("Loading labels of document %s", self.doc_id,)
        )
        promise = promise.then(lambda *args: None)  # drop logger return value

        promises = []
        self.core.call_all(
            "doc_get_labels_by_url_promise",
            promises, self.doc_url
        )
        for p in promises:
            promise = promise.then(p)

        return promise.then(self.show_labels)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 100

    def __init__(self):
        super().__init__()
        self.default_thumbnail = None
        self.running = False
        self.tasks = {}

        self.processing_docs = set()
        self.processed_docs = set()

    def get_interfaces(self):
        return [
            'gtk_doclist_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.maindow.doclist'],
            },
            {
                'interface': 'gtk_widget_label',
                'defaults': ['paperwork_gtk.widget.label'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all("work_queue_create", "labeler", stop_on_quit=True)

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def doclist_show(self, docs):
        self.core.call_all("work_queue_cancel_all", "labeler")
        self.task = {}

    def on_doc_list_clear(self):
        self.processed_docs = set()
        self.on_doc_list_visibility_changed()

    def on_doc_list_visibility_changed(self):
        self.core.call_all("work_queue_cancel_all", "labeler")
        self.processing_docs = set()
        self.nb_to_load = 0

    def on_doc_box_visible(self, doc_id, gtk_row, gtk_custom_flowlayout):
        if doc_id in self.processing_docs:
            return
        if doc_id in self.processed_docs:
            return

        self.processing_docs.add(doc_id)

        doc_url = self.core.call_success("doc_id_to_url", doc_id)
        if doc_url is None:
            return

        task = LabelingTask(self, doc_id, doc_url, gtk_custom_flowlayout)
        self.tasks[doc_url] = task

        def _when_loaded():
            if doc_id in self.processing_docs:
                self.processing_docs.remove(doc_id)
                self.processed_docs.add(doc_id)

        promise = task.get_promise()
        promise = promise.then(_when_loaded)

        self.core.call_success("work_queue_add_promise", "labeler", promise)

    def _refresh_doc(self, doc_url):
        if doc_url not in self.tasks:
            LOGGER.debug(
                "Labels on '%s' have changed, but it is not displayed at"
                " the moment", doc_url
            )
            return
        LOGGER.info("Reloading labels of '%s'", doc_url)
        self.core.call_success(
            "work_queue_add_promise", "labeler",
            self.tasks[doc_url].get_promise()
        )

    def doc_add_label_by_url(self, doc_url, label, color=None):
        self._refresh_doc(doc_url)

    def doc_remove_label_by_url(self, doc_url, label):
        self._refresh_doc(doc_url)



import openpaperwork_core


INSTRUCTIONS = {
    # 'Arch Linux / Manjaro / …': (
    #     "# TODO\n"
    # ),
    'Debian / Ubuntu / Mint / …': (
        "sudo apt install sane-utils\n"
        "sudo sh -c \"echo 127.0.0.1 >> /etc/sane.d/saned.conf\"\n"
        "sudo systemctl enable saned.socket\n"
        "sudo systemctl start saned.socket\n"
        "sudo adduser saned plugdev\n"
        "sudo adduser saned scanner\n"
        "sudo adduser saned lp\n"
        "# reboot\n"
        "\n"
        "# If your scanner is still not recognized, please check\n"
        "# https://gitlab.gnome.org/World/OpenPaperwork/paperwork/-/blob"
        "/develop/doc/install.flatpak.markdown#faq\n"
    ),
    'Fedora / CentOS / RHEL / …': (
        "sudo dnf install libinsane sane-backends-daemon\n"
        "sudo sh -c \"echo 127.0.0.1 >> /etc/sane.d/saned.conf\"\n"
        "sudo systemctl enable saned.socket\n"
        "sudo systemctl start saned.socket\n"
    ),
}


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.windows = []

    def get_interfaces(self):
        return [
            'gtk_settings_scanner_flatpak',
            'gtk_window_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'flatpak',
                'defaults': ['openpaperwork_core.flatpak'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_settings_scanner',
                'defaults': ['paperwork_gtk.settings.scanner.settings'],
            },
        ]

    def settings_scanner_get_extra_widget(self):
        if not self.core.call_success("is_in_flatpak"):
            return None

        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings.scanner",
            "flatpak.glade"
        )
        if widget_tree is None:
            return None
        button = widget_tree.get_object("button_flatpak_info")
        button.connect("clicked", self._on_clicked)
        return button

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def _on_clicked(self, button):
        widget_tree = self.core.call_success(
            "gtk_load_widget_tree", "paperwork_gtk.settings.scanner",
            "flatpak.glade"
        )
        if widget_tree is None:
            return

        dialog = widget_tree.get_object("flatpak_info_dialog")
        dialog.set_transient_for(self.windows[-1])
        dialog.set_modal(True)
        dialog.connect("response", self._on_close)
        dialog.connect("destroy", self._on_close)
        dialog.set_visible(True)

        selector = widget_tree.get_object("flatpak_info_selector")
        for k in sorted(INSTRUCTIONS.keys()):
            selector.append_text(k)

        selector.connect("changed", self._on_changed, widget_tree)
        selector.set_active(0)
        self._on_changed(selector, widget_tree)

    def _on_changed(self, selector, widget_tree):
        selected = selector.get_active_text()
        instruction = INSTRUCTIONS[selected]
        txt_buffer = widget_tree.get_object("textbuffer_instructions")
        txt_buffer.set_text(instruction)

    def _on_close(self, dialog, *args, **kwargs):
        dialog.set_visible(False)



import logging

import openpaperwork_core

from ... import _


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_delete"


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -100

    def get_interfaces(self):
        return [
            'menu',
            'menu_doc',
            'menu_doc_delete',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'action_doc_delete',
                'defaults': ['paperwork_gtk.actions.doc.delete'],
            },
            {
                'interface': 'doc_actions',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def on_doclist_initialized(self):
        self.core.call_all(
            "add_doc_action", _("Delete document"), "win." + ACTION_NAME
        )



try:
    # Fabulous is available and installed on GNU/Linux systems, but not on
    # Windows. Still this command can be called on Windows systems using
    # "paperwork-json"
    import fabulous
    import fabulous.color
    FABULOUS_AVAILABLE = True
except (ImportError, ValueError):
    FABULOUS_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.cmd.util

from .. import _


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return ['shell']

    def get_deps(self):
        return [
            {
                'interface': 'chkworkdir',
                'defaults': [
                    'paperwork_backend.chkworkdir.empty_doc',
                    'paperwork_backend.chkworkdir.label_color',
                ],
            },
        ]

    def cmd_complete_argparse(self, parser):
        p = parser.add_parser(
            'chkworkdir', help=_("Check and fix work directory integrity")
        )
        p.add_argument(
            '--yes', '-y', required=False, default=False, action='store_true',
            help=_("Don't ask to fix things, just fix them")
        )

    @staticmethod
    def _color(color):
        assert FABULOUS_AVAILABLE
        color = "#%1X%1X%1X" % (
            int(color[0] * 0xF),
            int(color[1] * 0xF),
            int(color[2] * 0xF),
        )
        return fabulous.color.bg256(color, "  ") + " "

    def cmd_run(self, console, args):
        if args.command != 'chkworkdir':
            return None

        console.print(_("Checking work directory ..."))

        problems = []
        self.core.call_all("check_work_dir", problems)

        if len(problems) <= 0:
            console.print(_("No problem found"))
            return problems

        console.print("")
        console.print(_("%d problems found:") % len(problems))
        for problem in problems:
            problem_color = (
                ""
                if "problem_color" not in problem
                else self._color(problem['problem_color'])
            )
            solution_color = (
                ""
                if "solution_color" not in problem
                else self._color(problem['solution_color'])
            )

            console.print("[{}]".format(problem['problem']))
            console.print(
                _("- Problem: ") + problem_color +
                problem['human_description']['problem']
            )
            console.print(
                _("- Possible solution: ") + solution_color +
                problem['human_description']['solution']
            )
            console.print("")

        if not args.yes:
            msg = _(
                "Do you want to fix those problems automatically"
                " using the indicated solutions ?"
            )
            r = openpaperwork_core.cmd.util.ask_confirmation(
                console,
                msg,
                default_interactive='n',
                default_non_interactive='n',
            )
            if r != 'y':
                console.print("OK, nothing changed.")
                return problems

        console.print(_("Fixing ..."))
        self.core.call_all("fix_work_dir", problems)
        console.print(_("All fixed !"))
        console.print(_("Synchronizing with work directory ..."))

        self.core.call_all("transaction_sync_all")
        self.core.call_all("mainloop_quit_graceful")
        self.core.call_one("mainloop")
        console.print(_("All done !"))

        return problems



"""
Takes of various document format that are actually converted to PDF files
so Paperwork can read them quickly
"""
import logging

import openpaperwork_core
import openpaperwork_core.promise

from .. import _


LOGGER = logging.Logger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 200

    def __init__(self):
        super().__init__()
        self.pending_doc_urls = set()
        self.file_types_by_ext = {}
        self.file_types_by_mime = {}
        self.cache_hash = {}

    def get_interfaces(self):
        return [
            "doc_convert_and_import",
            "doc_hash",
            "sync",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_converter',
                'defaults': ['paperwork_backend.converter.libreoffice'],
            },
            {
                'interface': 'doc_pdf_import',
                'defaults': ['paperwork_backend.model.pdf'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'transaction_manager',
                'defaults': ['paperwork_backend.sync'],
            },
        ]

    def init(self, core):
        super().init(core)

        file_types = []
        self.core.call_all("converter_get_file_types", file_types)

        self.file_types_by_ext = {}
        for (mime, ext, human_name) in file_types:  # order matters !
            if ext in self.file_types_by_ext:
                continue
            self.file_types_by_ext[ext] = mime

        self.file_types_by_mime = {}
        for (mime, ext, human_name) in file_types:  # order matters !
            if mime in self.file_types_by_ext:
                continue
            self.file_types_by_mime[mime] = ext

    def _is_converted(self, doc_url):
        if not self.core.call_success("fs_isdir", doc_url):
            return (None, None)
        for doc_file_url in self.core.call_success("fs_listdir", doc_url):
            doc_filename = self.core.call_success("fs_basename", doc_file_url)
            if "." not in doc_filename:
                continue
            (doc_filename, doc_ext) = doc_filename.rsplit(".", 1)
            doc_filename = doc_filename.lower()
            doc_ext = doc_ext.lower()
            if doc_filename == "doc" and doc_ext in self.file_types_by_ext:
                return (doc_file_url, doc_ext)
            return (None, None)
        else:
            # not a converted document
            return (None, None)

    def _update_pdf(self, doc_id, doc_url, doc_file_url, doc_ext):
        LOGGER.debug(
            "Document %s: %s (%s)",
            doc_id, doc_file_url, self.file_types_by_ext[doc_ext]
        )
        doc_mtime = self.core.call_success("fs_get_mtime", doc_file_url)

        pdf = self.core.call_success(
            "doc_get_pdf_url_by_url", doc_url, write=True
        )
        if not self.core.call_success("fs_exists", pdf):
            pdf_mtime = -1
        else:
            pdf_mtime = self.core.call_success("fs_get_mtime", pdf)

        if pdf_mtime >= doc_mtime:
            LOGGER.debug("Document %s: PDF is up-to-date")
            return False

        LOGGER.info(
            "Document %s: Updating PDF: %s --> %s",
            doc_id, doc_file_url, pdf
        )
        self.core.call_all(
            "on_progress", "converting", 0.0,
            _("Converting document %s to PDF ...") % self.core.call_success(
                "fs_basename", doc_file_url
            )
        )
        self.core.call_success(
            "convert_file_to_pdf",
            doc_file_url, self.file_types_by_ext[doc_ext],
            pdf
        )
        self.core.call_all("flush_doc_cache", doc_url)
        LOGGER.info("PDF updated")
        self.core.call_all("on_progress", "converting", 1.0)
        return True

    def page_get_img_url(self, doc_url, page_idx, write=False, **kwargs):
        if page_idx != 0:
            return
        # prevent double conversion
        if doc_url in self.pending_doc_urls:
            LOGGER.debug("Doc %s has already been checked", doc_url)
            return
        self.pending_doc_urls.add(doc_url)

        (doc_file_url, doc_ext) = self._is_converted(doc_url)
        if doc_file_url is None:
            self.pending_doc_urls.remove(doc_url)
            return

        doc_id = self.core.call_success("doc_url_to_id", doc_url)

        LOGGER.info("Checking conversion of document %s is up-to-date", doc_id)

        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self._update_pdf, args=(
                doc_id, doc_url, doc_file_url, doc_ext
            )
        )

        def do_transaction(has_changed):
            if not has_changed:
                return

            transactions = []
            self.core.call_all("doc_transaction_start", transactions, 1)
            transactions.sort(key=lambda t: -t.priority)

            try:
                for transaction in transactions:
                    transaction.del_doc(doc_id)
                for transaction in transactions:
                    transaction.commit()
            except Exception:
                for transaction in transactions:
                    transaction.cancel()

        promise = promise.then(openpaperwork_core.promise.ThreadedPromise(
            self.core, do_transaction
        ))
        promise = promise.then(self.pending_doc_urls.remove, doc_url)
        self.core.call_success("transaction_schedule", promise)

        return None

    def _check_all_docs(self):
        if len(self.file_types_by_ext) <= 0:
            LOGGER.info("No file converter available")
            return

        LOGGER.info("Checking all converted documents ...")

        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs, only_valid=False)

        self.pending_doc_urls.update((x[1] for x in all_docs))

        msg = _("Checking converted documents")
        self.core.call_all("on_progress", "converted_check", 0.0, msg)
        total = len(all_docs)
        for (idx, (doc_id, doc_url)) in enumerate(all_docs):
            (doc_file_url, doc_ext) = self._is_converted(doc_url)
            if doc_file_url is None:
                continue
            self._update_pdf(doc_id, doc_url, doc_file_url, doc_ext)
            self.pending_doc_urls.remove(doc_url)
            if idx % 100 == 0:
                self.core.call_all(
                    "on_progress", "converted_check", idx / total, msg
                )
        self.core.call_all("on_progress", "converted_check", 1.0)
        LOGGER.info("All converted documents have been checked")

    def sync(self, promises: list):
        promise = openpaperwork_core.promise.ThreadedPromise(
            self.core, self._check_all_docs
        )
        promises.insert(0, promise)

    def doc_convert_and_import(self, file_url):
        if "." in file_url:
            file_ext = file_url.rsplit(".", 1)[-1].lower()
        else:
            mime = self.core.call_success("fs_get_mime", file_url)
            if mime is not None:
                file_ext = self.file_types_by_mime[mime]
            else:
                LOGGER.error("Failed to figure out file type of '%s'", file_url)
                return (None, None)

        file_name = "doc." + file_ext
        (doc_id, doc_url) = self.core.call_success("storage_get_new_doc")
        dst_file_url = self.core.call_success("fs_join", doc_url, file_name)
        self.core.call_success("fs_mkdir_p", doc_url)
        try:
            self.core.call_success("fs_copy", file_url, dst_file_url)
            self._update_pdf(doc_id, doc_url, dst_file_url, file_ext)
            return (doc_id, doc_url)
        except Exception:
            self.core.call_success("fs_rm_rf", doc_url, trash=False)
            raise

    def doc_get_hash_by_url(self, doc_url):
        (doc_file_url, doc_ext) = self._is_converted(doc_url)
        if doc_file_url is None:
            return None
        if doc_url not in self.cache_hash:
            h = self.core.call_success("fs_hash", doc_file_url)
            self.cache_hash[doc_url] = h
        return self.cache_hash[doc_url]



import logging

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.widget_tree = None

    def get_interfaces(self):
        return [
            'gtk_scan_buttons_popover',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_scan_buttons',
                'defaults': [
                    'paperwork_gtk.mainwindow.docview.pageadd.buttons'
                ],
            },
        ]

    def init(self, core):
        super().init(core)

        opt = self.core.call_success(
            "config_build_simple", "pageadd", "active_source", lambda: None
        )
        self.core.call_all("config_register", "pageadd_active_source", opt)

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.pageadd", "source_popover.glade"
        )
        if self.widget_tree is None:
            # init must still work so 'chkdeps' is still available
            LOGGER.error("Failed to load widget tree")
            return

    def pageadd_sources_refresh(self):
        active = self.core.call_success("config_get", "pageadd_active_source")

        parent = self.widget_tree.get_object("page_sources_box")
        for child in parent.get_children():
            parent.remove(child)

        selectors = []
        self.core.call_all("pageadd_get_sources", selectors)
        for (selector, source_name, source_id, callback) in selectors[1:]:
            selector.join_group(selectors[0][0])
        for (selector, source_name, source_id, callback) in selectors:
            selector.connect(
                "toggled", self._on_toggle, source_name, source_id, callback
            )
            parent.pack_start(selector, expand=False, fill=True, padding=0)
        for (selector, source_name, source_id, callback) in selectors:
            if active == source_id:
                break
        else:
            active = None

        if active is None:
            if len(selectors) > 0:
                (selector, source_name, source_id, callback) = selectors[0]
                selector.set_active(True)
                self._on_toggle(selector, source_name, source_id, callback)
        else:
            for (selector, source_name, source_id, callback) in selectors:
                if active == source_id:
                    selector.set_active(True)
                    self._on_toggle(selector, source_name, source_id, callback)
                    break

        # if there is a single choice, there is no point in letting
        # the user choose. This is just confusing.
        self.core.call_all(
            "pageadd_buttons_set_source_popover",
            self.widget_tree.get_object("page_sources_popover")
            if len(selectors) > 1 else None
        )

    def _on_toggle(self, widget, source_name, source_id, callback):
        if not widget.get_active():
            return
        self.core.call_all(
            "pageadd_set_default_action", source_name, callback, source_id
        )
        self.core.call_all(
            "config_put", "pageadd_active_source", source_id
        )



import logging
import os

import openpaperwork_core


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 10000

    def get_deps(self):
        return [
            {
                'interface': 'flatpak',
                'defaults': ['openpaperwork_core.flatpak'],
            },
        ]

    def gtk_init(self, *args, **kwargs):
        # Flatpak workaround: make sure the fonts will be rendered correctly in
        # PDF files. Otherwise, sometime we get a weird error in the console:
        # "some font thing has failed" and no text is rendered when rendering
        # PDF pages.
        if self.core.call_success("is_in_flatpak"):
            LOGGER.info("Running `fc-cache -f` ...")
            r = os.system("fc-cache -f")
            r = os.waitstatus_to_exitcode(r)
            if r == 0:
                LOGGER.info("`fc-cache -f` has succeded")
            else:
                LOGGER.warning("`fc-cache -f` has failed: r=%d", r)



import openpaperwork_core
import openpaperwork_core.deps

GI_AVAILABLE = False
GLIB_AVAILABLE = False
POPPLER_AVAILABLE = False


try:
    import gi
    from gi.repository import GLib
    GI_AVAILABLE = True
except (ImportError, ValueError):
    pass


if GI_AVAILABLE:
    try:
        from gi.repository import Gio
        GLIB_AVAILABLE = True
    except (ImportError, ValueError):
        pass

    try:
        gi.require_version('Poppler', '0.18')
        from gi.repository import Poppler
        POPPLER_AVAILABLE = True
    except (ImportError, ValueError):
        pass


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -100

    def get_interfaces(self):
        return [
            'chkdeps',
            'poppler',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GI_AVAILABLE:
            out['gi'] = openpaperwork_core.deps.GI
        if not GLIB_AVAILABLE:
            out['glib'] = openpaperwork_core.deps.GLIB
        if not POPPLER_AVAILABLE:
            out['poppler'] = openpaperwork_core.deps.POPPLER

    def poppler_open(self, url, password=None):
        # Poppler.Document.new_from_data() expects .. a string
        # Poppler.Document.new_from_bytes() only exist starting with 0.82
        with self.core.call_success("fs_open", url, "rb") as fd:
            data = fd.read()
        ldata = len(data)
        data = GLib.Bytes.new(data)
        # Gio.MemoryInputStream.new_from_data() may leak
        # https://stackoverflow.com/questions/45838863/gio-memoryinputstream
        # --> use Gio.MemoryInputStream.new_from_bytes() instead
        data = Gio.MemoryInputStream.new_from_bytes(data)
        self.core.call_all("on_objref_track", data)
        doc = self.core.call_one(
            "mainloop_execute", Poppler.Document.new_from_stream,
            data, ldata, password=password
        )
        return doc



import logging

import openpaperwork_core
import openpaperwork_core.promise

from . import (
    DirectFileImporter,
    FileImport,
    RecursiveFileImporter
)
from .. import _


LOGGER = logging.getLogger(__name__)


class SinglePdfImporter(object):
    def __init__(self, core, file_import, src_file_uri, data):
        self.core = core
        self.file_import = file_import
        self.src_file_uri = src_file_uri
        self.data = data
        self.doc_id = None
        self.doc_url = None

    def _basic_import(self, file_uri):
        file_hash = self.core.call_success("fs_hash", file_uri)
        other_doc_id = self.core.call_success(
            "index_get_doc_id_by_hash", file_hash
        )
        if other_doc_id is not None:
            LOGGER.info("%s has already been imported", file_uri)
            self.file_import.stats[_("Already imported")] += 1
            return False

        LOGGER.info("Importing %s", file_uri)
        (self.doc_id, self.doc_url) = self.core.call_success(
            "doc_pdf_import", file_uri,
            password=self.data.get('password'),
            target_doc_id=self.file_import.active_doc_id
        )
        if self.doc_id is None:
            return False
        self.file_import.new_doc_ids.add(self.doc_id)
        self.file_import.stats[_("PDF")] += 1
        self.file_import.stats[_("Documents")] += 1
        return True

    def get_promise(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self._basic_import, args=(self.src_file_uri,)
        )


class SinglePdfImporterFactory(object):
    def __init__(self, plugin):
        self.plugin = plugin
        self.core = plugin.core

    @staticmethod
    def get_name():
        return _("Import PDF")

    @staticmethod
    def get_recursive_name():
        return _("Import PDFs recursively")

    def is_importable(self, core, file_uri):
        mime = core.call_success("fs_get_mime", file_uri)
        if mime is not None:
            if mime == Plugin.MIME_TYPE:
                return True
            return False
        if file_uri.lower().endswith(self.plugin.FILE_EXTENSION):
            return True

    def get_required_data(self, file_uri):
        try:
            self.core.call_success("poppler_open", file_uri, password=None)
            LOGGER.info("%s: not password protected", file_uri)
            return set()
        except Exception:
            # XXX(Jflesch): there is no specific exception type ... :/
            LOGGER.info("%s: password protected", file_uri)
            return {"password"}

    def make_importer(self, file_import, file_uri, data):
        return SinglePdfImporter(self.core, file_import, file_uri, data)


class Plugin(openpaperwork_core.PluginBase):
    FILE_EXTENSION = ".pdf"
    MIME_TYPE = "application/pdf"

    def get_interfaces(self):
        return [
            "import"
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_pdf_import',
                'defaults': ['paperwork_backend.model.pdf'],
            },
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_gtk.fs.gio'],
            },
            {
                'interface': 'mainloop',
                'defaults': ['openpaperwork_gtk.mainloop.glib'],
            },
            {
                'interface': 'thread',
                'defaults': ['openpaperwork_core.thread.simple'],
            },
        ]

    def get_import_mime_types(self, out: set):
        out.add(("PDF", self.MIME_TYPE))
        out.add((_("PDF folder"), "inode/directory"))

    def get_importer(self, out: list, file_import: FileImport):
        importer = DirectFileImporter(
            self.core, file_import, SinglePdfImporterFactory(self)
        )
        if importer.can_import():
            out.append(importer)

        importer = RecursiveFileImporter(
            self.core, file_import, SinglePdfImporterFactory(self)
        )
        if importer.can_import():
            out.append(importer)



import logging

try:
    from gi.repository import Gio
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)
ACTION_NAME = "doc_select_all"


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()
        self.visible_docs = []

    def get_interfaces(self):
        return [
            'action',
            'action_docs',
            'action_docs_select_all',
            'chkdeps',
            'search_listener',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'doc_selection',
                'defaults': ['paperwork_gtk.doc_selection'],
            },
        ]

    def init(self, core):
        super().init(core)
        if not GLIB_AVAILABLE:
            return
        action = Gio.SimpleAction.new(ACTION_NAME, None)
        action.connect("activate", self._select_all)
        self.core.call_all("app_actions_add", action)

    def chkdeps(self, out: dict):
        if not GLIB_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def on_search_results(self, query, docs):
        self.visible_docs = docs

    def _select_all(self, action, parameter):
        for doc in self.visible_docs:
            self.core.call_all("doc_selection_add", *doc)



"""
openpaperwork_core.beacon contains everything related to the communication
with the website https://www.openpaper.work/
"""

import datetime
import logging


LOGGER = logging.getLogger(__name__)


class PeriodicTask(object):
    def __init__(
                self, config_section_name, min_delay: datetime.timedelta,
                periodic_callback, else_callback=lambda: None
            ):
        self.config_section_name = config_section_name
        self.min_delay = min_delay
        self.periodic_callback = periodic_callback
        self.else_callback = else_callback

    def register_config(self, core):
        setting = core.call_success(
            "config_build_simple", self.config_section_name,
            "last_run", lambda: datetime.date(year=1970, month=1, day=1)
        )
        core.call_all(
            "config_register", self.config_section_name + "_last_run", setting
        )

    def do(self, core):
        now = datetime.date.today()
        last_run = core.call_success(
            "config_get", self.config_section_name + "_last_run"
        )
        if hasattr(last_run, 'value'):
            # ConfigDate object from
            # openpaperwork_core.config.backend.configparser
            last_run = last_run.value
        LOGGER.info(
            "[%s] Last run: %s ; Now: %s",
            self.config_section_name, last_run, now
        )
        if now - last_run < self.min_delay:
            LOGGER.info(
                "[%s] Nothing to do (%s < %s)",
                self.config_section_name, now - last_run, self.min_delay
            )
            self.else_callback()
            return

        LOGGER.info(
            "[%s] Running %s (%s >= %s)",
            self.config_section_name,
            self.periodic_callback, now - last_run, self.min_delay
        )
        self.periodic_callback()

        LOGGER.info("[%s] Updating last run date", self.config_section_name)
        core.call_all(
            "config_put", self.config_section_name + "_last_run", now
        )
        core.call_all("config_save")



import logging

import openpaperwork_core
import openpaperwork_core.deps

try:
    from gi.repository import GObject
    GLIB_AVAILABLE = True
except (ImportError, ValueError):
    GLIB_AVAILABLE = False

    # workaround so chkdeps can still be called
    class GObject(object):  # type: ignore[no-redef]
        TYPE_BOOLEAN = 0

        class SignalFlags(object):
            RUN_LAST = 0

        class GObject(object):
            pass


LOGGER = logging.getLogger(__name__)


class Scan(GObject.GObject):
    """
    Implements the same interface than the mainwindow.docview.pageview.Page
    objects, but display any scan relative to the current document.
    """

    __gsignals__ = {
        # never emitted
        'getting_size': (GObject.SignalFlags.RUN_LAST, None, ()),

        # emitted immediately and when a scan is started
        'size_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),

        # never emitted
        'img_obtained': (GObject.SignalFlags.RUN_LAST, None, ()),

        'visibility_changed': (GObject.SignalFlags.RUN_LAST, None, (
            # TODO(Jflesch): not implemented here (never emitted), but not
            # used anywhere anyway.
            GObject.TYPE_BOOLEAN,
        )),
    }

    def __init__(self, core, plugin, doc_id, page_idx, scan_id=None):
        super().__init__()
        self.core = core
        self.plugin = plugin
        self.doc_id = doc_id
        self.zoom = 1.0
        self.page_idx = page_idx

        self.scan_id = scan_id

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docview.scanview", "scanview.glade"
        )
        self.widget = self.widget_tree.get_object("scanview_area")
        self.visible = False

    def __str__(self):
        return "Scan renderer ({})".format(self.doc_id)

    def set_visible(self, visible):
        if visible == self.visible:
            return
        LOGGER.info("Scan renderer: Visible: {}".format(visible))
        self.visible = visible

    def start(self):
        LOGGER.info("Scan renderer: start")
        size = self.get_size()
        self.widget.set_size_request(size[0], size[1])
        self.widget.queue_resize()
        self.core.call_all("draw_scan_start", self.widget, self.scan_id)
        self.widget.set_visible(True)

    def stop(self):
        LOGGER.info("Scan renderer: stop")
        self.core.call_all("draw_scan_stop", self.widget)
        self.widget.set_visible(False)

    def get_visible(self):
        return self.visible

    def load(self):
        # the docview rely on this signal to know when pagss have been loaded
        # (in other words, when their size have been defined).
        # It remains in state 'loading' as long as not all the pages have
        # reported having be loaded --> we report immediately ourselves
        # as loaded
        self.core.call_all("on_page_size_obtained", self)
        self.emit("size_obtained")

    def close(self):
        pass

    def get_zoom(self):
        return self.zoom

    def set_zoom(self, zoom):
        self.zoom = zoom
        self.resize()

    def get_full_size(self):
        if self.scan_id is None:
            return (0, 0)
        size = self.core.call_success("draw_scan_get_scan_size", self.scan_id)
        if size is None:
            LOGGER.warning("Unknown scan_id: %s !?", self.scan_id)
            return (0, 0)
        return size

    def get_size(self):
        size = self.get_full_size()
        LOGGER.info("Scan size: %dx%d", size[0], size[1])
        return (
            int(size[0] * self.zoom),
            int(size[1] * self.zoom)
        )

    def resize(self):
        if self.scan_id is None:
            return
        size = self.get_size()
        self.widget.set_size_request(size[0], size[1])
        self.core.call_all("on_page_size_obtained", self)
        self.emit("size_obtained")
        self.widget.queue_resize()

    def refresh(self, reload=False):
        self.widget.queue_draw()

    def hide(self):
        pass

    def show(self):
        pass


if GLIB_AVAILABLE:
    GObject.type_register(Scan)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = -10000

    def __init__(self):
        super().__init__()
        self.scan = None
        self.doc_id = None
        self.doc_url = None

    def get_interfaces(self):
        return [
            'gtk_pageview',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'gtk_drawer_scan',
                'defaults': [
                    'openpaperwork_gtk.drawer.scan',
                    'paperwork_gtk.drawer.calibration',
                ],
            },
        ]

    def doc_open_components(self, out: list, doc_id, doc_url):
        self.doc_id = doc_id
        self.doc_url = doc_url

        nb_pages = self.core.call_success("doc_get_nb_pages_by_url", doc_url)
        if nb_pages is None:
            nb_pages = 0
        scan_id = self.core.call_success("scan2doc_doc_id_to_scan_id", doc_id)
        self.scan = Scan(self.core, self, doc_id, nb_pages, scan_id)
        if scan_id is None:
            out.append((False, self.scan))  # invisible
        else:
            out.append((True, self.scan))  # visible
            self.scan.start()

    def on_scan2doc_start(self, scan_id, doc_id, doc_url):
        if self.scan is None:
            return
        if doc_id != self.doc_id:
            return
        self.scan.scan_id = scan_id

    def on_scan_feed_start(self, scan_id):
        if self.scan is None:
            return
        if scan_id != self.scan.scan_id:
            return
        self.scan.start()
        self.core.call_all("docview_show_page_viewer", self.scan)

    def on_scan_page_start(self, scan_id, page_nb, scan_params):
        if self.scan is None:
            return
        if scan_id != self.scan.scan_id:
            return

        self.scan.resize()

        nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", self.doc_url
        )
        if nb_pages is None:
            nb_pages = 0

        handle_id = None

        def goto_page(*args, **kwargs):
            self.scan.widget.disconnect(handle_id)
            self.core.call_all("doc_goto_page", nb_pages)

        self.core.call_all("doc_goto_page", nb_pages)
        handle_id = self.scan.widget.connect("size-allocate", goto_page)

    def on_scan_chunk(self, scan_id, *args, **kwargs):
        self.scan.resize()

    def on_scan2doc_page_scanned(self, scan_id, doc_id, doc_url, page_idx):
        if self.scan is None:
            return
        if scan_id != self.scan.scan_id:
            return

        nb_pages = self.core.call_success(
            "doc_get_nb_pages_by_url", self.doc_url
        )
        assert nb_pages is not None

        LOGGER.info("Displaying new page %d", nb_pages - 1)
        self.core.call_all("doc_goto_page", nb_pages - 1)

    def on_scan_feed_end(self, scan_id):
        if self.scan is None:
            return
        if scan_id != self.scan.scan_id:
            return
        self.core.call_all("docview_hide_page_viewer", self.scan)
        self.scan.stop()

    def on_scan2doc_end(self, scan_id, doc_id, doc_url):
        if self.scan is None:
            return
        if doc_id != self.doc_id:
            return
        self.scan.scan_id = None



import glob
import locale
import logging
import os

import pycountry
import pyocr
import pyocr.builders

import openpaperwork_core

from . import util


LOGGER = logging.getLogger(__name__)

DEFAULT_OCR_LANG = "eng"  # if really we can't guess anything


def init_flatpak(core):
    """
    If we are in Flatpak, we must build a tessdata/ directory using the
    .traineddata files from each locale directory
    """
    tessdata_files = glob.glob("/app/share/locale/*/*.traineddata")
    if len(tessdata_files) <= 0:
        return

    paperwork_dir = core.call_success("paths_get_data_dir")
    tessdatadir = core.call_success("fs_join", paperwork_dir, "tessdata")
    tessdatadir = core.call_success("fs_unsafe", tessdatadir)

    LOGGER.info("Assuming we are running in Flatpak."
                " Building tessdata directory %s ...", tessdatadir)
    util.rm_rf(tessdatadir)
    os.makedirs(tessdatadir, exist_ok=True)

    os.symlink("/app/share/tessdata/eng.traineddata",
               os.path.join(tessdatadir, "eng.traineddata"))
    os.symlink("/app/share/tessdata/osd.traineddata",
               os.path.join(tessdatadir, "osd.traineddata"))
    os.symlink("/app/share/tessdata/configs",
               os.path.join(tessdatadir, "configs"))
    os.symlink("/app/share/tessdata/tessconfigs",
               os.path.join(tessdatadir, "tessconfigs"))
    for tessdata in tessdata_files:
        LOGGER.info("%s found", tessdata)
        os.symlink(
            tessdata, os.path.join(tessdatadir, os.path.basename(tessdata))
        )
    os.environ['TESSDATA_PREFIX'] = tessdatadir
    LOGGER.info("Tessdata directory ready")


def find_language(lang_str=None, allow_none=False):
    if lang_str is None:
        lang_str = locale.getdefaultlocale()[0]
        if lang_str == "C":
            LOGGER.warning("Locale is C. Assuming english !")
            return find_language(DEFAULT_OCR_LANG)
        if lang_str is None and not allow_none:
            LOGGER.warning("Unable to figure out locale. Assuming english !")
            return find_language(DEFAULT_OCR_LANG)
        if lang_str is None:
            LOGGER.warning("Unable to figure out locale !")
            return None

    lang_str = lang_str.lower()
    if "_" in lang_str:
        lang_str = lang_str.split("_")[0]
    LOGGER.info("System language: {}".format(lang_str))

    attrs = (
        'iso_639_3_code',
        'iso639_3_code',
        'iso639_2T_code',
        'iso639_1_code',
        'terminology',
        'bibliographic',
        'alpha_3',
        'alpha_2',
        'alpha2',
        'name',
    )
    for attr in attrs:
        try:
            r = pycountry.pycountry.languages.get(**{attr: lang_str})
            if r is not None:
                LOGGER.info("OCR language: {}".format(r))
                return r
        except (KeyError, UnicodeDecodeError):
            pass

    if allow_none:
        LOGGER.warning("Unknown language [{}]".format(lang_str))
        return None
    if lang_str is not None and lang_str == DEFAULT_OCR_LANG:
        raise Exception("Unable to find language !")
    LOGGER.warning("Unknown language [{}]. Switching back to english".format(
        lang_str
    ))
    return find_language(DEFAULT_OCR_LANG)


def pycountry_to_tesseract(ocr_langs, possibles=None):
    attrs = [
        'iso639_3_code',
        'terminology',
        'alpha_3',
    ]
    for attr in attrs:
        if not hasattr(ocr_langs, attr):
            continue
        if possibles is None or getattr(ocr_langs, attr) in possibles:
            r = getattr(ocr_langs, attr)
            if r is not None:
                return r
    return None


def get_default_ocr_langs(allow_none=False):
    # Try to guess based on the system locale what would be
    # the best OCR language

    ocr_tools = pyocr.get_available_tools()
    if len(ocr_tools) == 0:
        return None if allow_none else [DEFAULT_OCR_LANG]
    ocr_langs = ocr_tools[0].get_available_languages()

    lang = find_language(allow_none=True)
    if lang is None:
        return None if allow_none else [DEFAULT_OCR_LANG]
    lang = pycountry_to_tesseract(lang, ocr_langs)
    if lang is not None:
        return [lang]
    return None if allow_none else [DEFAULT_OCR_LANG]


class Plugin(openpaperwork_core.PluginBase):
    def __init__(self):
        super().__init__()

    def get_interfaces(self):
        return [
            "chkdeps",
            "ocr_settings",
        ]

    def get_deps(self):
        return [
            {
                'interface': 'config',
                'defaults': ['openpaperwork_core.config'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
            {
                'interface': 'data_versioning',
                'defaults': ['openpaperwork_core.data_versioning'],
            },
        ]

    def init(self, core):
        super().init(core)
        init_flatpak(self.core)

        ocr_langs = self.core.call_success(
            "config_build_simple",
            "OCR", "Lang", get_default_ocr_langs
        )
        self.core.call_all("config_register", "ocr_langs", ocr_langs)

    def chkdeps(self, out: dict):
        ocr_tools = pyocr.get_available_tools()
        if len(ocr_tools) <= 0:
            out['tesseract']['debian'] = 'tesseract-ocr'
            out['tesseract']['fedora'] = 'tesseract'
            out['tesseract']['gentoo'] = 'app-text/tesseract'
            out['tesseract']['linuxmint'] = 'tesseract-ocr'
            out['tesseract']['raspbian'] = 'tesseract-ocr'
            out['tesseract']['ubuntu'] = 'tesseract-ocr'
        ocr_lang = get_default_ocr_langs(allow_none=True)
        if ocr_lang is None:
            ocr_lang = find_language(allow_none=True)
            if ocr_lang is None:
                ocr_lang = "UNKNOWN"
            else:
                ocr_lang = pycountry_to_tesseract(ocr_lang)
                if ocr_lang is None:
                    ocr_lang = "<PYCOUNTRY ERROR>"
            name = 'tesseract-data-{}'.format(ocr_lang)
            out[name]['debian'] = 'tesseract-ocr-{}'.format(ocr_lang)
            out[name]['fedora'] = 'tesseract-langpack-{}'.format(ocr_lang)
            out[name]['linuxmint'] = 'tesseract-ocr-{}'.format(ocr_lang)
            out[name]['raspbian'] = 'tesseract-ocr-{}'.format(ocr_lang)
            out[name]['ubuntu'] = 'tesseract-ocr-{}'.format(ocr_lang)

    def ocr_get_active_langs(self):
        return self.core.call_success("config_get", "ocr_langs")

    def ocr_set_active_langs(self, langs):
        return self.core.call_success("config_put", "ocr_langs", langs)

    def ocr_is_enabled(self):
        if len(self.ocr_get_active_langs()) > 0:
            return True
        return None

    def ocr_add_observer_on_enabled(self, callback):
        self.core.call_all("config_add_observer", "ocr_langs", callback)

    def ocr_get_available_langs(self):
        ocr_tools = pyocr.get_available_tools()
        if len(ocr_tools) <= 0:
            return []
        return ocr_tools[0].get_available_languages()



import threading
import unittest

from .. import (Core, PluginBase)


class AbstractTestThread(unittest.TestCase):
    def get_plugin_name(self):
        """
        Must be overloaded by subclasses
        """
        assert False

    def setUp(self):
        class DummyMainloop(object):
            class Plugin(PluginBase):
                def get_interfaces(s):
                    return ['mainloop']

                def mainloop_ref(s, r):
                    pass

                def mainloop_unref(s, r):
                    pass

        self.core = Core(auto_load_dependencies=True)
        self.core._load_module("dummy_mainloop", DummyMainloop())
        self.core.load(self.get_plugin_name())
        self.core.init()

    def test_basic(self):
        out = {
            'task_a_done': False,
            'task_b_done': False,
        }
        sem = threading.Semaphore(value=0)

        def task_a():
            out['task_a_done'] = True
            sem.release()

        def task_b():
            out['task_b_done'] = True
            sem.release()

        self.core.call_all("on_mainloop_start")

        self.core.call_one("thread_start", task_a)
        self.core.call_one("thread_start", task_b)

        for _ in range(0, 2):
            sem.acquire()

        self.assertTrue(out['task_a_done'])
        self.assertTrue(out['task_b_done'])

        self.core.call_all("on_mainloop_quit")

    def test_mainloop_restart(self):
        # mainloop can be stopped and started again many times
        out = {}
        sem = threading.Semaphore(value=0)

        def task_a():
            out['task_a_done'] = True
            sem.release()

        out['task_a_done'] = False
        self.core.call_all("on_mainloop_start")
        self.core.call_one("thread_start", task_a)
        for _ in range(0, 1):
            sem.acquire()
        self.assertTrue(out['task_a_done'])
        self.core.call_all("on_mainloop_quit")

        out['task_a_done'] = False
        self.core.call_all("on_mainloop_start")
        self.core.call_one("thread_start", task_a)
        for _ in range(0, 1):
            sem.acquire()
        self.assertTrue(out['task_a_done'])
        self.core.call_all("on_mainloop_quit")



import datetime
import logging
import time

import openpaperwork_core
import openpaperwork_core.promise


LOGGER = logging.getLogger(__name__)


class BaseTransaction(object):
    def __init__(self, core, total_expected):
        self.core = core
        self.processed = 0
        self.total = total_expected
        self._current_doc = None
        self._current_doc_pages = -1

    def notify_progress(
            self, upd_type, description, page_nb=-1, total_pages=-1):
        if self.total <= self.processed:
            self.total = self.processed + 1
        if self.total <= 0:
            progression = 0
        else:
            progression = self.processed / self.total

        if page_nb >= 0 and total_pages > 0:
            progression += (page_nb / total_pages / self.total)

        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_progress", upd_type, progression, description
        )

    def notify_done(self, upd_type):
        self.core.call_one(
            "mainloop_schedule", self.core.call_all,
            "on_progress", upd_type, 1.0
        )

    def add_doc(self, doc_id):
        self._current_doc = doc_id
        self._current_doc_pages = -1
        self.processed += 1

    def upd_doc(self, doc_id):
        self._current_doc = doc_id
        self._current_doc_pages = -1
        self.processed += 1

    def del_doc(self, doc_id):
        self._current_doc = doc_id
        self._current_doc_pages = -1
        self.processed += 1

    def unchanged_doc(self, doc_id):
        self.processed += 1

    def cancel(self):
        self._current_doc = None
        self._current_doc_pages = -1

    def commit(self):
        self._current_doc = None


def diff_lists(list_old, list_new):
    """
    Returns a dictionary giving the differences between both lists.

    Objects in the list must have 2 attributes:
    - `key`
    - `extra`

    `key` values identify each object in both lists.
    Comparisons are done on `key` + `extra`.

    `list_old` will be unrolled immediately.

    Arguments:
        `list_old` -- [a, b, c, ...]
        `list_new` -- [a, c_modified, d, ...]

    Returns:
        [
            ('same', a.key),
            ('del', b.key),
            ('upd', c.key),
            ('add', d.key),
        ]

        list_old + diff => list_new
    """
    list_old = {obj.key: obj for obj in list_old}
    examined_new = set()

    for obj_new in list_new:
        examined_new.add(obj_new.key)

        if obj_new.key not in list_old:
            yield ('added', obj_new.key)
            continue

        obj_old = list_old[obj_new.key]
        if obj_new.extra != obj_old.extra:
            yield ('updated', obj_new.key)
        else:
            yield ('unchanged', obj_new.key)

    for obj_old in list_old.values():
        if obj_old.key not in examined_new:
            yield ('deleted', obj_old.key)


class StorageDoc(object):
    def __init__(self, core, doc_id, doc_url):
        self.core = core
        self.key = doc_id
        self.doc_url = doc_url

    def get_mtime(self):
        mtime = self.core.call_success("doc_get_mtime_by_url", self.doc_url)
        if mtime is None:
            mtime = 0
        return datetime.datetime.fromtimestamp(mtime)

    extra = property(get_mtime)


class Syncer(object):
    """
    This object allows to compare mtimes progressively. It then calls
    the methods `add_doc`, `del_doc`, `upd_doc` and `commit` on the given
    transaction object.

    Useful to handle calls to 'sync' (see interface 'syncable').
    """
    def __init__(self, core, names, new_all, old_all, transactions):
        self.core = core
        self.names = names
        self.new_all = new_all
        self.old_all = old_all
        self.transactions = transactions
        self.diff_generator = None
        self.start = None
        self.nb_compared = 0

        self.work_dir_id = self.core.call_success("storage_get_id")

    def get_promise(self):
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self.run
        )

    def run(self):
        self.start = time.time()
        self.diff_generator = diff_lists(self.old_all, self.new_all)

        try:
            for (action, key) in self.diff_generator:

                current_work_dir = self.core.call_success("storage_get_id")
                if self.work_dir_id != current_work_dir:
                    LOGGER.warning(
                        "%s: Work directory has changed during sync."
                        " Cancelling transactions",
                        self.names
                    )
                    for transaction in self.transactions:
                        transaction.cancel()
                    return

                self.nb_compared += 1
                if action != 'unchanged':
                    LOGGER.info("Sync: %s, %s", action, key)
                    for name in self.names:
                        self.core.call_one(
                            "mainloop_schedule", self.core.call_all,
                            "on_sync", name, action, key
                        )
                if action == "added":
                    for transaction in self.transactions:
                        LOGGER.debug(
                            "transaction.add_doc<%s>(%s)",
                            transaction.add_doc, key
                        )
                        transaction.add_doc(key)
                elif action == "updated":
                    for transaction in self.transactions:
                        LOGGER.debug(
                            "transaction.upd_doc<%s>(%s)",
                            transaction.upd_doc, key
                        )
                        transaction.upd_doc(key)
                elif action == "deleted":
                    for transaction in self.transactions:
                        LOGGER.debug(
                            "transaction.del_doc<%s>(%s)",
                            transaction.del_doc, key
                        )
                        transaction.del_doc(key)
                else:
                    for transaction in self.transactions:
                        LOGGER.debug(
                            "transaction.unchanged_doc<%s>(%s)",
                            transaction.unchanged_doc, key
                        )
                        transaction.unchanged_doc(key)

            LOGGER.info("Sync: Committing ...")
            for transaction in self.transactions:
                transaction.commit()
            LOGGER.info("Sync: Committed")
            stop = time.time()
            LOGGER.info(
                "%s: Has compared %d objects in %.3fs",
                self.names, self.nb_compared, stop - self.start
            )
        except Exception as exc:
            LOGGER.error(
                "%s: Fail to sync. Cancelling transactions",
                self.names, exc_info=exc
            )
            for transaction in self.transactions:
                transaction.cancel()


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 20000

    def get_interfaces(self):
        return ['transaction_manager']

    def get_deps(self):
        return [
            {
                'interface': 'work_queue',
                'defaults': ['openpaperwork_core.work_queue.default'],
            },
        ]

    def init(self, core):
        super().init(core)
        self.core.call_all("work_queue_create", "transactions")

    def transaction_schedule(self, promise):
        """
        Transactions should never be run in parrallel (even if on the same
        thread). Some databases (Sqlite3) don't support that.
        --> we use a work queue to ensure they are run one after the other.
        """
        return self.core.call_success(
            "work_queue_add_promise", "transactions", promise
        )

    def _transaction_simple(self, changes):
        if len(changes) <= 0:
            LOGGER.info("No change. Nothing to do")
            return

        transactions = []
        self.core.call_all("doc_transaction_start", transactions, len(changes))
        transactions.sort(key=lambda transaction: -transaction.priority)

        try:
            for (change, doc_id) in changes:
                doc_url = self.core.call_success("doc_id_to_url", doc_id)
                if doc_url is None:
                    change = 'del'
                elif self.core.call_success("is_doc", doc_url) is None:
                    change = 'del'
                for transaction in transactions:
                    if change == 'add':
                        transaction.add_doc(doc_id)
                    elif change == 'upd':
                        transaction.upd_doc(doc_id)
                    elif change == 'del':
                        transaction.del_doc(doc_id)
                    else:
                        raise Exception("Unknown change type: %s" % change)

            for transaction in transactions:
                transaction.commit()
        except Exception as exc:
            LOGGER.error("Transactions have failed", exc_info=exc)
            for transaction in transactions:
                transaction.cancel()
            raise

    def transaction_simple_promise(self, changes):
        """
        See transaction_simple(). Must be scheduled with
        'transaction_schedule()'.
        """
        return openpaperwork_core.promise.ThreadedPromise(
            self.core, self._transaction_simple, args=(changes,)
        )

    def transaction_simple(self, changes: list):
        """
        Helper method.
        Schedules a transaction for a bunch of document ids.

        Changes must be a list:
        [
            ('add', 'some_doc_id'),
            ('upd', 'some_doc_id_2'),
            ('upd', 'some_doc_id_3'),
            ('del', 'some_doc_id_4'),
        ]
        """
        return self.transaction_schedule(
            self.core.call_success("transaction_simple_promise", changes)
        )

    def transaction_sync_all(self):
        """
        Make sure all the plugins synchronize their databases with the work
        directory.
        """
        promises = []
        self.core.call_all("sync", promises)
        promise = promises[0]
        for p in promises[1:]:
            promise = promise.then(p)
        self.transaction_schedule(promise)



import enum
import logging
import re

try:
    import gi
    gi.require_version('Gdk', '3.0')
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gdk
    from gi.repository import Gtk
    GTK_AVAILABLE = True
except (ImportError, ValueError):
    GTK_AVAILABLE = False

import openpaperwork_core
import openpaperwork_gtk.deps
import paperwork_backend.sync

from ... import _


LOGGER = logging.getLogger(__name__)

# forbid:
# - empty strings
# - strings that contain a comma
RE_FORBIDDEN_LABELS = re.compile("(^$|.*,.*)")


class LabelChange(enum.Enum):
    UNCHANGED = 0
    ADDED = 1
    REMOVED = 2


class LabelAction(object):
    def __init__(self, current_state):
        # current_state == None for multiple document selection
        self.current_state = current_state
        self.change = LabelChange.UNCHANGED

    def on_change(self):
        if self.current_state is None:

            # multi-docs mode : rotate over the 3 possible change values
            if self.change == LabelChange.UNCHANGED:
                self.change = LabelChange.ADDED
            elif self.change == LabelChange.ADDED:
                self.change = LabelChange.REMOVED
            elif self.change == LabelChange.REMOVED:
                self.change = LabelChange.UNCHANGED

        else:

            # single-doc mode

            if not self.current_state:

                if self.change == LabelChange.UNCHANGED:
                    self.change = LabelChange.ADDED
                else:
                    self.change = LabelChange.UNCHANGED

            else:

                if self.change == LabelChange.UNCHANGED:
                    self.change = LabelChange.REMOVED
                else:
                    self.change = LabelChange.UNCHANGED

    def get_image(self):
        if self.change == LabelChange.ADDED:
            return "list-add-symbolic"
        elif self.change == LabelChange.REMOVED:
            return "list-remove-symbolic"
        else:
            if self.current_state:
                return "object-select-symbolic"
            return None

    def __str__(self):
        return str((self.change, self.current_state))


class LabelEditor(object):
    def __init__(self, plugin):
        super().__init__()
        self.core = plugin.core
        self.plugin = plugin

        self.active_docs = []

        # self.changed_labels contains the label that have been modified
        # (color or text)
        self.changed_labels = {}

        # self.toggled_labels indicates the labels that have been selected or
        # unselected (value = LabelAction, depending on whether they are
        # now selected or not)
        self.toggled_labels = {}

        # self.new_labels contains the labels that user has just created
        # (but are not yet applied on any document)
        self.new_labels = set()

        # self.deleted_labels contains the labels that the user wants to
        # remove from *all* the documents.
        self.deleted_labels = set()

        self.widget_tree = self.core.call_success(
            "gtk_load_widget_tree",
            "paperwork_gtk.mainwindow.docproperties", "labels.glade"
        )
        color_widget = self.widget_tree.get_object("new_label_color")
        color_widget.set_rgba(self.core.call_success(
            "gtk_theme_get_color", "theme_bg_color"
        ))

        self.widget_tree.get_object("new_label_button").connect(
            "clicked", self._on_new_label
        )
        new_label_entry = self.widget_tree.get_object("new_label_entry")
        new_label_entry.connect("activate", self._on_new_label)
        new_label_entry.connect("changed", self._on_label_txt_changed)

    def _update_button_img(self, toggle, label_action):
        img = label_action.get_image()
        if img is not None:
            image = Gtk.Image.new_from_icon_name(img, Gtk.IconSize.MENU)
        else:
            image = Gtk.Image()
        toggle.set_image(image)

    def _refresh_list(self):
        listbox = self.widget_tree.get_object("listbox_labels")
        for widget in list(listbox.get_children()):
            listbox.remove(widget)

        doc_labels = None
        if len(self.active_docs) == 1:
            doc_labels = set()
            active_doc = list(self.active_docs)[0]
            self.core.call_all(
                "doc_get_labels_by_url", doc_labels, active_doc[1]
            )

        labels = set()
        self.core.call_all("labels_get_all", labels)
        if doc_labels is not None:
            labels.update(doc_labels)

        labels = [
            (
                self.core.call_success("i18n_strip_accents", label[0].lower()),
                label[0],
                label[1]
            ) for label in labels
        ]
        labels.sort()
        labels = [(label[1], label[2]) for label in labels]

        if doc_labels is not None:
            doc_labels = {doc_label[0] for doc_label in doc_labels}
        for old_label in labels:
            if old_label in self.changed_labels:
                new_label = self.changed_labels[old_label]
            else:
                new_label = old_label

            if old_label in self.deleted_labels:
                continue

            if old_label in self.toggled_labels:
                label_action = self.toggled_labels[old_label]
            else:
                if doc_labels is None:
                    active = None
                else:
                    active = old_label[0] in doc_labels
                label_action = LabelAction(active)

            widget_tree_label = self.core.call_success(
                "gtk_load_widget_tree",
                "paperwork_gtk.mainwindow.docproperties", "label.glade"
            )
            button = widget_tree_label.get_object("label_label")
            button.set_label(new_label[0])
            button.connect("clicked", self._on_label_button_clicked, old_label)

            toggle = widget_tree_label.get_object("toggle_button")
            toggle.set_sensitive(True)
            self._update_button_img(toggle, label_action)
            toggle.connect("clicked", self._on_toggle, old_label, label_action)

            color = self.core.call_success("label_color_to_rgb", new_label[1])
            color_button = widget_tree_label.get_object("color_button")
            color_button.set_sensitive(old_label not in self.new_labels)
            color_button.set_rgba(Gdk.RGBA(color[0], color[1], color[2], 1.0))
            color_button.connect(
                "color-set", self._on_color_changed, old_label
            )

            delete = widget_tree_label.get_object("delete_button")
            delete.connect("clicked", self._on_delete, old_label)

            listbox.add(widget_tree_label.get_object("label_row"))

        new_labels = list(self.new_labels)
        new_labels.sort()
        for label in new_labels:
            widget_tree_label = self.core.call_success(
                "gtk_load_widget_tree",
                "paperwork_gtk.mainwindow.docproperties", "label.glade"
            )
            button = widget_tree_label.get_object("label_label")
            button.set_label(label[0])
            button.set_sensitive(False)

            toggle = widget_tree_label.get_object("toggle_button")
            toggle.set_sensitive(False)
            image = Gtk.Image.new_from_icon_name(
                "list-add-symbolic", Gtk.IconSize.MENU
            )
            toggle.set_image(image)

            color = self.core.call_success("label_color_to_rgb", label[1])
            color_button = widget_tree_label.get_object("color_button")
            color_button.set_sensitive(True)
            color_button.set_rgba(Gdk.RGBA(color[0], color[1], color[2], 1.0))
            color_button.connect(
                "color-set", self._on_new_color_changed, label
            )

            delete = widget_tree_label.get_object("delete_button")
            delete.connect("clicked", self._on_new_delete, label)

            listbox.add(widget_tree_label.get_object("label_row"))

        listbox.add(self.widget_tree.get_object("row_add_label"))

    def _on_label_button_clicked(self, label_button, label):
        original_label = label
        if label in self.changed_labels:
            label = self.changed_labels[label]
        self.core.call_success(
            "gtk_show_dialog_single_entry",
            self.plugin, _("Renaming label"),
            label[0], self, label_button, original_label, label
        )

    def on_label_rename_response(
            self, new_label_txt, label_button, original_label, label):
        new_label_txt = new_label_txt.strip()

        if not self._check_label_name(new_label_txt):
            return

        LOGGER.info("Renaming label %s --> %s", label, new_label_txt)
        self.changed_labels[original_label] = (new_label_txt, label[1])

        label_button.set_label(new_label_txt)

    def _on_toggle(self, button, label, label_action):
        label_action.on_change()
        self._update_button_img(button, label_action)
        self.toggled_labels[label] = label_action

    def _on_color_changed(self, button, original_label):
        color = button.get_rgba()
        color = (color.red, color.green, color.blue)
        color = self.core.call_success("label_color_from_rgb", color)

        label = original_label
        if original_label in self.changed_labels:
            label = self.changed_labels[original_label]
        new_label = (label[0], color)

        self.changed_labels[original_label] = new_label

    def _on_new_color_changed(self, button, original_label):
        self.new_labels.remove(original_label)
        color = button.get_rgba()
        color = (color.red, color.green, color.blue)
        color = self.core.call_success("label_color_from_rgb", color)
        new_label = (original_label[0], color)
        self.new_labels.add(new_label)
        self._refresh_list()

    def _on_delete(self, button, original_label):
        self.core.call_all(
            "gtk_show_dialog_yes_no",
            self,
            _(
                "Are you sure you want to delete label '%s'"
                " from ALL documents ?"
            ) % (original_label,),
            original_label,
        )

    def on_dialog_yes_no_reply(self, origin, response, *args, **kwargs):
        if origin is not self:
            return
        if not response:
            LOGGER.info("Label delete canceled")
            return
        (original_label,) = args
        LOGGER.info("Will delete label %s on all documents", original_label)
        self.deleted_labels.add(original_label)
        self._refresh_list()

    def _on_new_delete(self, button, label):
        self.new_labels.remove(label)
        self._refresh_list()

    def _check_label_name(self, label_name):
        all_labels = set()
        self.core.call_all("labels_get_all", all_labels)
        for label in set(all_labels):
            if label in self.changed_labels:
                all_labels.add(self.changed_labels[label])
        all_labels.update(self.new_labels)
        all_labels = {label[0] for label in all_labels}

        if label_name in all_labels:
            return False

        if RE_FORBIDDEN_LABELS.match(label_name):
            return False

        return True

    def _on_label_txt_changed(self, *args, **kwargs):
        entry = self.widget_tree.get_object("new_label_entry")
        button = self.widget_tree.get_object("new_label_button")
        txt = entry.get_text().strip()

        valid = self._check_label_name(txt)

        button.set_sensitive(valid)
        if valid or txt == "":
            self.core.call_all("gtk_entry_reset_colors", entry)
        else:
            self.core.call_all("gtk_entry_set_colors", entry)

    def _on_new_label(self, *args, **kwargs):
        text = self.widget_tree.get_object("new_label_entry").get_text()
        text = text.strip()
        if text == "":
            LOGGER.info("New label requested, but no text provided")
            return

        color_widget = self.widget_tree.get_object("new_label_color")

        color = color_widget.get_rgba()
        color = (color.red, color.green, color.blue)
        color = self.core.call_success("label_color_from_rgb", color)

        self.new_labels.add((text, color))
        self._refresh_list()

        # reset fields
        self.widget_tree.get_object("new_label_entry").set_text("")
        color_widget.set_rgba(self.core.call_success(
            "gtk_theme_get_color", "theme_bg_color"
        ))

    def _make_new_labels(self, out, doc_id, doc_url):
        if len(self.new_labels) <= 0:
            return
        for label in self.new_labels:
            self.core.call_success(
                "doc_add_label_by_url", doc_url, label[0], label[1]
            )
        out.upd_docs.add(doc_id)

    def _make_label_updates(self, out):
        if len(self.changed_labels) <= 0:
            return

        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs)
        all_docs = set(all_docs)

        current = 0
        total = len(self.changed_labels) * len(all_docs)

        for (doc_id, doc_url) in all_docs:
            for (old_label, new_label) in self.changed_labels.items():
                LOGGER.info(
                    "Changing label %s into %s on document %s",
                    old_label, new_label, doc_id
                )
                self.core.call_all(
                    "on_progress", "label_upd", current / total,
                    _(
                        "Changing label {old_label} into {new_label}"
                        " on document {doc_id}"
                    ).format(
                        old_label=old_label, new_label=new_label, doc_id=doc_id
                    )
                )
                current += 1
                doc_labels = set()
                self.core.call_all(
                    "doc_get_labels_by_url", doc_labels, doc_url
                )
                if old_label not in doc_labels:
                    continue
                out.upd_docs.add(doc_id)
                self.core.call_all(
                    "doc_remove_label_by_url", doc_url, old_label[0]
                )
                self.core.call_success(
                    "doc_add_label_by_url", doc_url,
                    new_label[0], new_label[1]
                )
        self.core.call_all("on_progress", "label_upd", 1.0)

    def _make_label_toggling(self, out, doc_id, doc_url):
        if len(self.toggled_labels) <= 0:
            return

        for (label, action) in self.toggled_labels.items():
            doc_labels = set()
            self.core.call_all("doc_get_labels_by_url", doc_labels, doc_url)

            if action.change == LabelChange.ADDED:
                if label in doc_labels:
                    continue
                self.core.call_success(
                    "doc_add_label_by_url", doc_url, label[0], label[1]
                )
                out.upd_docs.add(doc_id)
            elif action.change == LabelChange.REMOVED:
                if label not in doc_labels:
                    continue
                self.core.call_all(
                    "doc_remove_label_by_url", doc_url, label[0]
                )
                out.upd_docs.add(doc_id)

    def _make_label_deletions(self, out):
        if len(self.deleted_labels) <= 0:
            return

        all_docs = []
        self.core.call_all("storage_get_all_docs", all_docs)
        all_docs = set(all_docs)

        current = 0
        total = len(self.deleted_labels) * len(all_docs)

        for (doc_id, doc_url) in all_docs:
            for old_label in self.deleted_labels:
                LOGGER.info(
                    "Deleting label %s from document %s",
                    old_label, doc_id
                )
                self.core.call_all(
                    "on_progress", "label_del", current / total,
                    _(
                        "Deleting label {old_label} from document {doc_id}"
                    ).format(old_label=old_label, doc_id=doc_id)
                )
                current += 1
                doc_labels = set()
                self.core.call_all(
                    "doc_get_labels_by_url", doc_labels, doc_url
                )
                if old_label not in doc_labels:
                    continue
                out.upd_docs.add(doc_id)
                self.core.call_all(
                    "doc_remove_label_by_url", doc_url, old_label[0]
                )
        self.core.call_all("on_progress", "label_del", 1.0)

    def doc_properties_components_set_active_doc(self, doc_id, doc_url):
        self.doc_properties_components_set_active_docs({(doc_id, doc_url)})

    def doc_properties_components_set_active_docs(self, docs: set):
        self.active_docs = docs
        self.changed_labels = {}
        self.toggled_labels = {}
        self.new_labels = set()
        self.deleted_labels = set()
        new_label_entry = self.widget_tree.get_object("new_label_entry")
        new_label_entry.set_text("")

        self._refresh_list()

    def doc_properties_components_apply_changes(self, out):
        LOGGER.info("Selected/Unselected labels: %s", self.toggled_labels)
        LOGGER.info("Modified labels: %s", self.changed_labels)
        LOGGER.info("New labels: %s", self.new_labels)
        LOGGER.info("Deleted labels: %s", self.deleted_labels)

        if len(self.active_docs) == 1:
            # The document may have been renamed: use out.doc_id instead of
            # self.active_docs
            doc_id = out.doc_id
            doc_url = self.core.call_success("doc_id_to_url", doc_id)
            self.active_docs = {(doc_id, doc_url)}

        for (doc_id, doc_url) in self.active_docs:
            LOGGER.info("Updating document %s", doc_id)
            self._make_label_toggling(out, doc_id, doc_url)
            self._make_new_labels(out, doc_id, doc_url)

        if len(self.changed_labels) <= 0 and len(self.deleted_labels) <= 0:
            return
        self._make_label_deletions(out)
        self._make_label_updates(out)

    def doc_properties_components_cancel_changes(self):
        self.doc_properties_components_set_active_docs(self.active_docs)


class Plugin(openpaperwork_core.PluginBase):
    PRIORITY = 1000

    def __init__(self):
        super().__init__()
        self.windows = []
        self.editors = []

    def get_interfaces(self):
        return [
            'chkdeps',
            'gtk_doc_property',
            'gtk_window_listener',
            'screenshot_provider',
            'syncable',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'doc_labels',
                'defaults': ['paperwork_backend.model.labels'],
            },
            {
                'interface': 'document_storage',
                'defaults': ['paperwork_backend.model.workdir'],
            },
            {
                'interface': 'gtk_colors',
                'defaults': ['openpaperwork_gtk.colors'],
            },
            {
                'interface': 'gtk_dialog_yes_no',
                'defaults': ['openpaperwork_gtk.dialogs.yes_no'],
            },
            {
                'interface': 'gtk_doc_properties',
                'defaults': ['paperwork_gtk.docproperties'],
            },
            {
                'interface': 'gtk_resources',
                'defaults': ['openpaperwork_gtk.resources'],
            },
            {
                'interface': 'i18n',
                'defaults': ['openpaperwork_core.i18n.python'],
            },
            {
                'interface': 'screenshot',
                'defaults': ['openpaperwork_gtk.screenshots'],
            },
        ]

    def on_gtk_window_opened(self, window):
        self.windows.append(window)

    def on_gtk_window_closed(self, window):
        self.windows.remove(window)

    def init(self, core):
        super().init(core)
        if not GTK_AVAILABLE:
            # chkdeps() must still be callable
            return

    def chkdeps(self, out: dict):
        if not GTK_AVAILABLE:
            out['gtk'].update(openpaperwork_gtk.deps.GTK)

    def doc_properties_components_get(self, out: list, multiple_docs=False):
        editor = LabelEditor(self)
        self.editors.append(editor)
        out.append(editor.widget_tree.get_object("listbox_global"))

    def on_dialog_yes_no_reply(self, origin, response, *args, **kwargs):
        for editor in self.editors:
            editor.on_dialog_yes_no_reply(origin, response, *args, **kwargs)

    def doc_properties_components_set_active_doc(self, doc_id, doc_url):
        for editor in self.editors:
            editor.doc_properties_components_set_active_doc(doc_id, doc_url)

    def doc_properties_components_set_active_docs(self, docs: set):
        for editor in self.editors:
            editor.doc_properties_components_set_active_docs(docs)

    def doc_properties_components_apply_changes(self, out):
        for editor in self.editors:
            editor.doc_properties_components_apply_changes(out)

    def doc_properties_components_cancel_changes(self):
        for editor in self.editors:
            editor.doc_properties_components_cancel_changes()

    def on_dialog_single_entry_reply(
            self, parent, r, new_value, *args, **kwargs):
        if parent is not self:
            return
        if not r:
            return
        editor = args[0]
        args = args[1:]
        editor.on_label_rename_response(new_value, *args, **kwargs)

    def doc_transaction_start(self, out: list, total_expected=-1):
        class RefreshTransaction(paperwork_backend.sync.BaseTransaction):
            priority = -100000

            def commit(s):
                for editor in self.editors:
                    self.core.call_one(
                        "mainloop_schedule", editor._refresh_list
                    )

        out.append(RefreshTransaction(self.core, total_expected))

    def sync(self, promises: list):
        pass

    def on_label_loading_end(self):
        for editor in self.editors:
            editor._refresh_list()

    def screenshot_snap_all_doc_widgets(self, out_dir):
        for editor in self.editors:
            self.core.call_success(
                "screenshot_snap_widget",
                editor.widget_tree.get_object("listbox_global"),
                self.core.call_success(
                    "fs_join", out_dir, "doc_labels.png"
                ),
                margins=(10, 10, 10, -400)
            )
            self.core.call_success(
                "screenshot_snap_widget",
                editor.widget_tree.get_object("new_label_button"),
                self.core.call_success(
                    "fs_join", out_dir, "doc_new_label.png"
                ),
                margins=(400, 50, 50, 50)
            )



import logging

try:
    from gi.repository import Gio
    GIO_AVAILABLE = True
except (ImportError, ValueError):
    GIO_AVAILABLE = False

import openpaperwork_core
import openpaperwork_core.deps


LOGGER = logging.getLogger(__name__)


class Plugin(openpaperwork_core.PluginBase):
    def get_interfaces(self):
        return [
            'chkdeps',
            'action',
            'action_doc',
            'action_doc_prev_next',
        ]

    def get_deps(self):
        return [
            {
                'interface': 'app_actions',
                'defaults': ['paperwork_gtk.mainwindow.window'],
            },
            {
                'interface': 'gtk_doclist',
                'defaults': ['paperwork_gtk.mainwindow.doclist'],
            },
        ]

    def chkdeps(self, out: dict):
        if not GIO_AVAILABLE:
            out['glib'].update(openpaperwork_core.deps.GLIB)

    def init(self, core):
        super().init(core)
        if not GIO_AVAILABLE:
            return
        action = Gio.SimpleAction.new('doc_prev', None)
        action.connect("activate", self._goto, -1)
        self.core.call_all("app_actions_add", action)
        action = Gio.SimpleAction.new('doc_next', None)
        action.connect("activate", self._goto, 1)
        self.core.call_all("app_actions_add", action)

    def _goto(self, action, parameter, offset):
        self.core.call_all("open_next_doc", offset)



import logging

import openpaperwork_core


LOGGER = logging.getLogger(__name__)

# TODO(Jflesch): this version shouldn't be common to all applications using
# Openpaperwork-core
DATA_VERSION = 2


class Plugin(openpaperwork_core.PluginBase):
    """
    Keep a version number in ~/.local/share/paperwork2/data_version.
    If the version doesn't match 'DATA_VERSION', right before syncing,
    delete ~/.local/share/paperwork2 to force a full synchronisation.
    """

    def get_interfaces(self):
        return ['data_versioning']

    def get_deps(self):
        return [
            {
                'interface': 'fs',
                'defaults': ['openpaperwork_core.fs.python'],
            },
            {
                'interface': 'paths',
                'defaults': ['openpaperwork_core.paths.xdg'],
            },
        ]

    def init(self, core):
        super().init(core)

        data_dir = self.core.call_success("paths_get_data_dir")
        data_ver_file = self.core.call_success(
            "fs_join", data_dir, "data_version"
        )

        data_version = -1
        if self.core.call_success("fs_exists", data_ver_file):
            with self.core.call_success("fs_open", data_ver_file, 'r') as fd:
                data_version = int(fd.read().strip())

        LOGGER.info(
            "Expected data version: %d ; Current data version: %d",
            DATA_VERSION, data_version
        )

        if data_version != DATA_VERSION:
            LOGGER.warning(
                "Data version doesn't match (%d != %d). Forcing full sync",
                data_version, DATA_VERSION
            )

            self.core.call_success("fs_rm_rf", data_dir, trash=False)
            # call 'paths_get_data_dir' again to recreate the data directory
            data_dir = self.core.call_success("paths_get_data_dir")

            with self.core.call_success("fs_open", data_ver_file, 'w') as fd:
                fd.write(str(DATA_VERSION))

            self.core.call_all("on_data_files_deleted")



from . import PluginBase


class Plugin(PluginBase):
    """
    Plugin implementing the interface 'app' just provide some very basic
    information regarding the application we are building.
    """

    def get_interfaces(self):
        return ['app']

    def app_get_name(self):
        return "OpenPaperwork Core"

    def app_get_fs_name(self):
        return "openpaperwork_core"

    def app_get_version(self):
        return "0.0"



import logging
import threading
import traceback


LOGGER = logging.getLogger(__name__)


class BasePromise(object):
    def __init__(
                self, core, func=None, args=None, kwargs=None, parent=None,
                hide_caught_exceptions=False
            ):
        # we allow dummy Promise with not function provided. It allows to
        # write cleaner code in some cases.
        self.core = core
        self.func = func
        self.hide_caught_exceptions = hide_caught_exceptions
        if args is None:
            self.args = ()
        else:
            self.args = args
        if kwargs is None:
            self.kwargs = {}
        else:
            self.kwargs = kwargs
        self.parent = parent

        self._then = []
        self._catch = []

        self.scheduled = False

        self.parent_promise_return = None
        self.created_by = traceback.extract_stack()

    def __str__(self):
        return "Promise<{}>({})".format(str(self.func), id(self))

    def __repr__(self):
        return str(self)

    def then(self, callback, *args, **kwargs):
        if isinstance(callback, BasePromise):
            assert args is None or len(args) <= 0
            assert kwargs is None or len(kwargs) <= 0
            last_promise = callback
            while callback.parent is not None:
                callback = callback.parent
            next_promise = callback
            next_promise.parent = self
        else:
            next_promise = Promise(
                self.core, callback, args, kwargs, parent=self
            )
            last_promise = next_promise
        if not next_promise.hide_caught_exceptions:
            next_promise.hide_caught_exceptions = self.hide_caught_exceptions
        self._then.append(next_promise)
        return last_promise

    def catch(self, callback, *args, **kwargs):
        self._catch.append((callback, args, kwargs))
        return self

    def on_error(self, exc, hide_caught_exceptions=False):
        self.scheduled = False
        hide_caught_exceptions = (
            hide_caught_exceptions or self.hide_caught_exceptions
        )
        if len(self._catch) > 0:
            if hide_caught_exceptions:
                trace = lambda *args, **kwargs: None  # NOQA: E731
            else:
                trace = LOGGER.warning
            caught = "caught"
        elif len(self._then) > 0:
            for t in self._then:
                t.scheduled = False
                t.on_error(exc, hide_caught_exceptions)
            return
        else:
            trace = LOGGER.error
            caught = "uncaught"

        trace("=== %s exception in promise ===", caught, exc_info=exc)
        trace("promise.func=%s", self.func)
        trace("promise.args=%s", self.args)
        trace("promise.kwargs=%s", self.kwargs)
        trace("promise.parent=%s", self.parent)
        trace(
            "promise.parent_promise_return=%s", self.parent_promise_return
        )
        trace("=== Promise was created by ===")
        for (idx, stack_el) in enumerate(self.created_by):
            trace(
                "%2d: %20s: L%5d: %s",
                idx, stack_el[0], stack_el[1], stack_el[2]
            )

        if len(self._catch) > 0:
            for (c, args, kwargs) in self._catch:
                self.core.call_one(
                    "mainloop_schedule", c, exc, *args, **kwargs
                )
            return

        raise exc

    def _do(self, args):
        try:
            self.do(args)
        finally:
            self.scheduled = False

    def schedule(self, *args):
        scheduled = self.scheduled

        s = self
        while s.parent is not None:
            scheduled = scheduled or s.scheduled
            s.scheduled = True
            s = s.parent
        s.scheduled = True

        if scheduled:
            # a parent promise already scheduled --> we made sure to mark
            # all the children promises as scheduled, but there is no
            # need to actually schedule the parent.
            return

        if args == ():
            args = None
        self.core.call_one("mainloop_schedule", s._do, args)

    def wait(self):
        assert (
            # must never be called from main loop.
            threading.current_thread().ident !=
            self.core.call_success("mainloop_get_thread_id")
        )
        event = threading.Event()
        out = None

        def wakeup(r=None):
            nonlocal out
            out = r

        self.then(event.set)
        event.wait()
        return out


class Promise(BasePromise):
    """
    Executed in the main loop thread.
    Requires a plugin implementing the interface 'mainloop'.
    """

    def do(self, parent_r=None):
        self.parent_promise_return = parent_r
        try:
            if self.func is None:
                our_r = None
            else:
                if parent_r is None:
                    args = self.args
                else:
                    args = (parent_r,) + self.args

                LOGGER.debug(
                    "Promise: Begin: %s(%s, %s)", self.func, args, self.kwargs
                )
                our_r = self.func(*args, **self.kwargs)
                LOGGER.debug(
                    "Promise: End: %s(%s, %s)", self.func, args, self.kwargs
                )

            for t in self._then:
                self.core.call_one("mainloop_schedule", t._do, our_r)
        except Exception as exc:
            self.on_error(exc)


class DelayPromise(BasePromise):
    """
    Promise adding delay between 2 other promises.
    Requires a plugin implementing the interface 'mainloop'
    """
    def __init__(self, core, delay_s):
        super().__init__(core)
        self.delay_s = delay_s

    def __str__(self):
        return "DelayPromise<delay_s={}>({})".format(self.delay_s, id(self))

    def _call_then(self, parent_r):
        for t in self._then:
            self.core.call_one("mainloop_schedule", t._do, parent_r)

    def do(self, parent_r=None):
        LOGGER.debug("Promise: delay: %fs", self.delay_s)
        self.core.call_one(
            "mainloop_schedule", self._call_then, parent_r,
            delay_s=self.delay_s
        )


class ThreadedPromise(BasePromise):
    """
    Promise for which the provided callback will be run in another thread,
    leaving the main loop thread free to do other things.
    Requires a plugin implementing the interface 'mainloop' and a plugin
    implementing the interface 'thread'.

    IMPORTANT: This should ONLY be used for long-lasting tasks that cannot
    be split in small tasks (image processing, OCR, etc). The callback provided
    must be really careful regarding thread-safety.
    """

    def __str__(self):
        return "ThreadedPromise<{}>({})".format(str(self.func), id(self))

    def _threaded_do(self, parent_r):
        try:
            if parent_r is None:
                args = self.args
            else:
                args = (parent_r,) + self.args

            LOGGER.debug(
                "Threaded promise: Begin: %s(%s, %s)",
                self.func, args, self.kwargs
            )
            our_r = self.func(*args, **self.kwargs)
            LOGGER.debug(
                "Threaded promise: end: %s(%s, %s)",
                self.func, args, self.kwargs
            )

            for t in self._then:
                self.core.call_one("mainloop_schedule", t._do, our_r)
        except Exception as exc:
            self.core.call_one("mainloop_schedule", self.on_error, exc)

    def do(self, parent_r=None):
        self.parent_promise_return = parent_r
        try:
            if self.func is None:
                # no thread, we immediately schedule the next promises
                for t in self._then:
                    self.core.call_one("mainloop_schedule", t._do, None)
                return

            self.core.call_one("thread_start", self._threaded_do, parent_r)
        except Exception as exc:
            self.on_error(exc)
            return
